
def genai_safety(threshold: str = "BLOCK_ONLY_HIGH"):
    """
    BLOCK_ONLY_HIGH - block when high probability of unsafe content is detected
    BLOCK_MEDIUM_AND_ABOVE - block when medium or high probability of content is detected
    BLOCK_LOW_AND_ABOVE - block when low, medium, or high probability of unsafe content is detected
    BLOCK_NONE - no block, but need to be on an allow list to use
    """
    from google.generativeai.types import (
        HarmCategory,
        HarmBlockThreshold
    )

    if threshold == 'BLOCK_ONLY_HIGH':
        thresh = HarmBlockThreshold.BLOCK_ONLY_HIGH
    elif threshold == 'BLOCK_MEDIUM_AND_ABOVE':
        thresh = HarmBlockThreshold.BLOCK_MEDIUM_AND_ABOVE
    elif threshold == 'BLOCK_LOW_AND_ABOVE':
        thresh = HarmBlockThreshold.BLOCK_LOW_AND_ABOVE
    elif threshold == 'BLOCK_NONE':
        thresh = HarmBlockThreshold.BLOCK_NONE
    else:
        raise ValueError("Invalid threshold")

    safety_settings = {
        HarmCategory.HARM_CATEGORY_HARASSMENT: thresh,
        HarmCategory.HARM_CATEGORY_HATE_SPEECH: thresh,
        HarmCategory.HARM_CATEGORY_SEXUALLY_EXPLICIT: thresh,
        HarmCategory.HARM_CATEGORY_DANGEROUS_CONTENT: thresh,
    }

    return safety_settings


#   Copyright [2024] [Holosun ApS]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import datetime
import os
import base64
import uuid
import time

try:
    from google.cloud import storage
except ImportError:
    storage = None

from ..custom_logging import log
from ..utils import ConfigManager
from ..utils.mime import guess_mime_type


def handle_base64_image(base64_data: str, vector_name: str, extension: str):
    """
    Handle base64 image data, decode it, save it as a file, upload it to GCS, and return the image URI and MIME type.

    Args:
        base64_data (str): The base64 encoded image data.
        vector_name (str): The vector name for the GCS path.
        extension (str): The file extension of the image (e.g., ".jpg", ".png").

    Returns:
        Tuple[str, str]: The URI of the uploaded image and the MIME type.
    """
    
    model = ConfigManager(vector_name).vacConfig("llm")
    if "," not in base64_data:
        raise ValueError(f"Invalid Base64 image data: Missing metadata header {base64_data}")

    if model.startswith("openai"):  # pass it to gpt directly
        mime_type = base64_data.split(",", 1)
        log.info(f"Returning directly as openai - mime_type: {mime_type}")
        return base64_data, mime_type

    try:
        header, encoded = base64_data.split(",", 1)
        data = base64.b64decode(encoded)

        filename = f"{str(uuid.uuid4())}{extension}"
        with open(filename, "wb") as f:
            f.write(data)

        image_uri = add_file_to_gcs(filename, vector_name)
        os.remove(filename)  # Clean up the saved file

        # Determine MIME type based on filename
        mime_type = guess_mime_type(filename) or "application/octet-stream"

        return image_uri, mime_type
    except Exception as e:
        raise Exception(f'Base64 image upload failed: {str(e)}')


def resolve_bucket(vector_name):
    bucket_name = None

    if os.getenv('EXTENSIONS_BUCKET'):
        log.warning('Resolving to EXTENSIONS_BUCKET environment variable')
        return os.getenv('EXTENSIONS_BUCKET')
    
    if vector_name:
        bucket_config = ConfigManager(vector_name).vacConfig("upload")

        if bucket_config:
            if bucket_config.get("buckets"):
                bucket_name = bucket_config.get("buckets").get("all")

    bucket_name = bucket_name or os.getenv('GCS_BUCKET')
    if not bucket_name:
        raise ValueError("No bucket found to upload to: GCS_BUCKET returned None")
    
    if bucket_name.startswith("gs://"):
        bucket_name = bucket_name.removeprefix("gs://")
    
    return bucket_name

def add_folder_to_gcs(
        source_folder:str, 
        vector_name:str=None, 
        bucket_name:str=None, 
        metadata:dict=None, 
        bucket_folderpath:str=None):
    """Uploads a folder and all its contents to a specified GCS bucket."""

    uris = []
    for root, dirs, files in os.walk(source_folder):
        for file in files:
            local_path = os.path.join(root, file)
            # Create the relative path in the destination folder
            relative_path = os.path.relpath(local_path, source_folder)
            bucket_filepath = os.path.join(bucket_folderpath, relative_path)

            uri = add_file_to_gcs(local_path, 
                            vector_name=vector_name, 
                            bucket_name=bucket_name, 
                            metadata=metadata, 
                            bucket_filepath=bucket_filepath)
            uris.append(uri)
            
    log.info(f"uploaded [{len(files)}] to GCS bucket")

    return uris

def add_file_to_gcs(filename: str, 
                    vector_name:str=None, 
                    bucket_name: str=None, 
                    metadata:dict=None, 
                    bucket_filepath:str=None):

    if not storage:
        return None
    
    try:
        storage_client = storage.Client()
    except Exception as err:
        log.error(f"Error creating storage client: {str(err)}")
        return None
    
    if not bucket_name:
        bucket_name = resolve_bucket(vector_name)
    
    bucket = storage_client.get_bucket(bucket_name)
    now = datetime.datetime.now()
    year = now.strftime("%Y")
    month = now.strftime("%m")
    day = now.strftime("%d") 
    hour = now.strftime("%H")
    hour_prev = (now - datetime.timedelta(hours=1)).strftime("%H")

    if os.getenv('EXTENSIONS_BUCKET'):
        log.warning(f"setting {bucket_filepath=} to {os.path.basename(filename)} basename due to EXTENSIONS_BUCKET setting")
        bucket_filepath = os.path.basename(filename)

    if not vector_name:
        vector_name = "global"
    
    if not bucket_filepath:
        
        bucket_filepath = f"{vector_name}/{year}/{month}/{day}/{hour}/{os.path.basename(filename)}"
    
    bucket_filepath_prev = f"{vector_name}/{year}/{month}/{day}/{hour_prev}/{os.path.basename(filename)}"

    blob = bucket.blob(bucket_filepath)
    blob_prev = bucket.blob(bucket_filepath_prev)

    if blob.exists() and not os.getenv('EXTENSIONS_BUCKET'):
        log.info(f"File {filename} already exists in gs://{bucket_name}/{bucket_filepath}")
        return f"gs://{bucket_name}/{bucket_filepath}"

    if blob_prev.exists():
        log.info(f"File {filename} already exists in gs://{bucket_name}/{bucket_filepath_prev}")
        return f"gs://{bucket_name}/{bucket_filepath_prev}"

    log.debug(f"File {filename} does not already exist in bucket {bucket_name}/{bucket_filepath}")

    the_metadata = {
        "vector_name": vector_name,
    }
    if metadata is not None:
        the_metadata.update(metadata)

    blob.metadata = the_metadata
    
    max_retries = 5
    base_delay = 1  # 1 second
    for attempt in range(max_retries):
        try:
            blob.upload_from_filename(filename)
            log.info(f"File {filename} uploaded to gs://{bucket_name}/{bucket_filepath}")
            break  # Success! Exit the loop.
        except Exception as e:
            # In case of an exception (timeout, etc.), wait and then retry
            log.warning(f"Upload attempt {attempt + 1} failed with error: {str(e)}. Retrying...")
            time.sleep(base_delay * (2 ** attempt))  # Exponential backoff

    else:  # This block executes if the loop completes without breaking
        log.error(f"Failed to upload file {filename} to gs://{bucket_name}/{bucket_filepath} after {max_retries} attempts.")

    return f"gs://{bucket_name}/{bucket_filepath}"

def get_pdf_split_file_name(object_id, part_name):
    # Get the base file name without the file extension and directory
    base_name = os.path.basename(object_id).rsplit('.', 1)[0]

    # Return the full object name for the image
    return f"{os.path.dirname(object_id)}/{base_name}/pdf_parts/{part_name}"

def get_summary_file_name(object_id):
    # Get the base file name without the file extension and directory
    base_name = os.path.basename(object_id).rsplit('.', 1)[0]

    # Return the full object name for the image
    return f"{os.path.dirname(object_id)}/{base_name}/summary.md"    

def get_image_file_name(object_id, image_name, mime_type):
    # Get the base file name without the file extension and directory
    base_name = os.path.basename(object_id).rsplit('.', 1)[0]
    # Define a mapping from MIME types to file extensions
    file_extension_mapping = {
        "image/jpeg": "jpeg",
        "image/png": "png",
        "image/gif": "gif",
        # Add other MIME types and extensions as needed
    }
    # Get the file extension for the given mime type
    file_extension = file_extension_mapping.get(mime_type, "jpeg")
    # Return the full object name for the image
    return f"{os.path.dirname(object_id)}/{base_name}/img/{image_name}.{file_extension}"




import re
from ..utils.mime import guess_mime_type
from ..gcs import get_bytes_from_gcs
from ..custom_logging import log
import os
import tempfile
try:
    import google.generativeai as genai
except ImportError:
    genai = None

def extract_gs_images_and_genai_upload(content: str, limit:int=20):
    # Regular expression to find gs:// URLs 
    pattern = r'gs://[^ ]+\.(?:png|jpg|jpeg|pdf)'

    gs_matches = re.findall(pattern, content)
    # only 20 images by default
    unique_gs_matches = list(set(gs_matches))[:limit] 
    output_gs_images = []
    
    for gs_uri in unique_gs_matches:
        mime_type = guess_mime_type(gs_uri)
        if mime_type is None:
            continue
        
        log.info(f"Getting bytes from GCS: {gs_uri}")
        image_bytes = get_bytes_from_gcs(gs_uri)
        if image_bytes is None:
            continue

        # Get the basename from the gs_uri to use as the file name
        file_name = os.path.basename(gs_uri)

        # Create a temporary directory and write the file with the basename
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_file_path = os.path.join(temp_dir, file_name)
            
            # Write the BytesIO object to the file
            with open(temp_file_path, 'wb') as temp_file:
                temp_file.write(image_bytes)

            # Pass the temporary file's path to the upload function
            try:
                uploaded_file = genai.upload_file(temp_file_path)
                output_gs_images.append(uploaded_file)
            except Exception as e:
                log.error(f"Error adding {gs_uri} to base64: {str(e)}")
    
    return output_gs_images



import json
import traceback
import datetime
import uuid
import random
from functools import partial
import inspect
import asyncio
from typing import Dict, List, Optional, Callable, Any


from ..chat_history import extract_chat_history_with_cache, extract_chat_history_async_cached
from ...qna.parsers import parse_output
from ...streaming import start_streaming_chat, start_streaming_chat_async
from ...archive import archive_qa
from ...custom_logging import log
from ...utils import ConfigManager
from ...utils.version import sunholo_version
import os
from ...gcs.add_file import add_file_to_gcs, handle_base64_image
from ..swagger import validate_api_key
from datetime import timedelta

try:
    from flask import request, jsonify, Response
except ImportError:
    pass 

try:
    from ...pubsub import PubSubManager
except ImportError:
    PubSubManager = None

try:
    from ...mcp.mcp_manager import MCPClientManager
except ImportError:
    MCPClientManager = None

try:
    from ...mcp.vac_mcp_server import VACMCPServer
    from mcp.server import Server
    from mcp.types import JSONRPCMessage, ErrorData, INTERNAL_ERROR
except ImportError:
    VACMCPServer = None
    Server = None
    JSONRPCMessage = None
    ErrorData = None
    INTERNAL_ERROR = None

try:
    from ...a2a.vac_a2a_agent import VACA2AAgent
except (ImportError, SyntaxError):
    VACA2AAgent = None


# Cache dictionary to store validated API keys
api_key_cache = {}
cache_duration = timedelta(minutes=5)  # Cache duration

class VACRoutes:
    """
**Usage Example:**

```python
from agents.flask import VACRoutes

app = Flask(__name__)

def stream_interpreter(question, vector_name, chat_history, **kwargs):
    # Implement your streaming logic
    ...

def vac_interpreter(question, vector_name, chat_history, **kwargs):
    # Implement your static VAC logic
    ...

vac_routes = VACRoutes(app, stream_interpreter, vac_interpreter)

if __name__ == "__main__":
    app.run(debug=True)
```
    
    """
    def __init__(self, app, 
                 stream_interpreter: callable, 
                 vac_interpreter:callable=None, 
                 additional_routes:dict=None, 
                 mcp_servers: List[Dict[str, Any]] = None,
                 async_stream:bool=False,
                 add_langfuse_eval:bool=True,
                 enable_mcp_server:bool=False,
                 enable_a2a_agent:bool=False,
                 a2a_vac_names: List[str] = None):
        self.app = app
        self.stream_interpreter = stream_interpreter
        self.vac_interpreter = vac_interpreter or partial(self.vac_interpreter_default)

        # MCP client initialization
        self.mcp_servers = mcp_servers or []
        self.mcp_client_manager = MCPClientManager()    
        # Initialize MCP connections
        if self.mcp_servers and self.mcp_client_manager:
            asyncio.create_task(self._initialize_mcp_servers())

        # MCP server initialization
        self.enable_mcp_server = enable_mcp_server
        self.vac_mcp_server = None
        if self.enable_mcp_server and VACMCPServer:
            self.vac_mcp_server = VACMCPServer(
                stream_interpreter=self.stream_interpreter,
                vac_interpreter=self.vac_interpreter
            )

        # A2A agent initialization  
        self.enable_a2a_agent = enable_a2a_agent
        self.vac_a2a_agent = None
        self.a2a_vac_names = a2a_vac_names
        if self.enable_a2a_agent and VACA2AAgent:
            # Extract base URL from request context during route handling
            # For now, initialize with placeholder - will be updated in route handlers
            self.vac_a2a_agent = None  # Initialized lazily in route handlers

        self.additional_routes = additional_routes if additional_routes is not None else []
        self.async_stream = async_stream
        self.add_langfuse_eval = add_langfuse_eval
        self.register_routes()
        

    def vac_interpreter_default(self, question: str, vector_name: str, chat_history=[], **kwargs):
        # Create a callback that does nothing for streaming if you don't want intermediate outputs
        class NoOpCallback:
            def on_llm_new_token(self, token):
                pass
            def on_llm_end(self, response):
                pass

        # Use the NoOpCallback for non-streaming behavior
        callback = NoOpCallback()

        # Pass all arguments to vac_stream and use the final return
        result = self.stream_interpreter(
            question=question, 
            vector_name=vector_name, 
            chat_history=chat_history, 
            callback=callback, 
            **kwargs
        )

        return result

    def register_routes(self):
        """
        Registers all the VAC routes for the Flask application.
        """
        # Basic routes
        self.app.route("/", methods=['GET'])(self.home)
        self.app.route("/health", methods=['GET'])(self.health)
        
        if self.async_stream:  # Use async treatment
            log.info("async_stream enabled")
            self.app.route('/vac/streaming/<vector_name>', 
                        methods=['POST'], 
                        provide_automatic_options=False)(self.handle_stream_vac_async)
        else:
            self.app.route('/vac/streaming/<vector_name>', 
                        methods=['POST'], 
                        provide_automatic_options=False)(self.handle_stream_vac)
        # Static VAC
        self.app.route('/vac/<vector_name>', 
                       methods=['POST'], 
                       provide_automatic_options=False)(self.handle_process_vac)

        # Authentication middleware
        self.app.before_request(self.check_authentication)

        # Handle OPTIONS requests explicitly
        self.app.route('/vac/streaming/<vector_name>', methods=['OPTIONS'])(self.handle_options)
        self.app.route('/vac/<vector_name>', methods=['OPTIONS'])(self.handle_options)

        # OpenAI health endpoint
        self.app.route('/openai/health', methods=['GET', 'POST'])(self.openai_health_endpoint)

        # OpenAI compatible endpoint
        self.app.route('/openai/v1/chat/completions', methods=['POST'])(self.handle_openai_compatible_endpoint)
        self.app.route('/openai/v1/chat/completions/<vector_name>', methods=['POST'])(self.handle_openai_compatible_endpoint)

        # MCP client routes
        if self.mcp_servers:
            self.app.route('/mcp/tools', methods=['GET'])(self.handle_mcp_list_tools)
            self.app.route('/mcp/tools/<server_name>', methods=['GET'])(self.handle_mcp_list_tools)
            self.app.route('/mcp/call', methods=['POST'])(self.handle_mcp_call_tool)
            self.app.route('/mcp/resources', methods=['GET'])(self.handle_mcp_list_resources)
            self.app.route('/mcp/resources/read', methods=['POST'])(self.handle_mcp_read_resource)
        
        # MCP server endpoint
        if self.enable_mcp_server and self.vac_mcp_server:
            self.app.route('/mcp', methods=['POST', 'GET'])(self.handle_mcp_server)
        
        # A2A agent endpoints
        if self.enable_a2a_agent:
            self.app.route('/.well-known/agent.json', methods=['GET'])(self.handle_a2a_agent_card)
            self.app.route('/a2a/tasks/send', methods=['POST'])(self.handle_a2a_task_send)
            self.app.route('/a2a/tasks/sendSubscribe', methods=['POST'])(self.handle_a2a_task_send_subscribe)
            self.app.route('/a2a/tasks/get', methods=['POST'])(self.handle_a2a_task_get)
            self.app.route('/a2a/tasks/cancel', methods=['POST'])(self.handle_a2a_task_cancel)
            self.app.route('/a2a/tasks/pushNotification/set', methods=['POST'])(self.handle_a2a_push_notification)
    
        self.register_additional_routes()

        self.app.after_request(self.register_after_request)

    def handle_options(self, **kwargs):
        response = Response(status=200)
        response.headers['Access-Control-Allow-Origin'] = '*'
        
        # Reflect the request's `Access-Control-Request-Headers`
        request_headers = request.headers.get('Access-Control-Request-Headers', '')
        response.headers.add('Access-Control-Allow-Headers', request_headers)

        # Specify allowed methods
        response.headers.add('Access-Control-Allow-Methods', 'GET, POST, OPTIONS')
        log.info(f'OPTION Request headers: {request_headers}')

        return response

    def register_after_request(self, response):
        # Ensure correct CORS headers
        response.headers['Access-Control-Allow-Origin'] = '*'
        response.headers['Access-Control-Allow-Headers'] = 'Content-Type, Authorization, x-api-key'
        response.headers['Access-Control-Allow-Methods'] = 'GET, POST, OPTIONS'
        
        return response

    def register_additional_routes(self):
        """
        Registers additional custom routes provided during initialization.

        Example:
        ```python
        from flask import Flask, jsonify
        from agents.flask import VACRoutes

        app = Flask(__name__)

        def stream_interpreter(question, vector_name, chat_history, **kwargs):
            # Implement your streaming logic
            ...

        def vac_interpreter(question, vector_name, chat_history, **kwargs):
            # Implement your static VAC logic
            ...

        def custom_handler():
            return jsonify({"message": "Custom route!"})

        custom_routes = [
            {
                "rule": "/custom",
                "methods": ["GET"],
                "handler": custom_handler
            }
        ]

        vac_routes = VACRoutes(app, stream_interpreter, vac_interpreter, additional_routes=custom_routes)

        if __name__ == "__main__":
            app.run(debug=True)
        ```
        """
        for route in self.additional_routes:
            self.app.route(route["rule"], methods=route["methods"])(route["handler"])

    def home(self):
        return jsonify("OK")

    def health(self):
        return jsonify({"status": "healthy"})
    
    def make_openai_response(self, user_message, vector_name, answer):
        response_id = str(uuid.uuid4())
        log.info("openai response: Q: {user_message} to VECTOR_NAME: {vector_name} - A: {answer}")
        openai_response = {
            "id": response_id,
            "object": "chat.completion",
            "created": str(int(datetime.datetime.now().timestamp())),
            "model": vector_name,
            "system_fingerprint": sunholo_version(),
            "choices": [{
                "index": 0,
                "message": {
                    "role": "assistant",
                    "content": answer,
                },
                "logprobs": None,
                "finish_reason": "stop"
            }],
            "usage": {
                "prompt_tokens": len(user_message.split()),
                "completion_tokens": len(answer.split()),
                "total_tokens": len(user_message.split()) + len(answer.split())
            }
        }

        log.info(f"OpenAI response: {openai_response}")
        return jsonify(openai_response)
        
    def handle_stream_vac(self, vector_name):
        observed_stream_interpreter = self.stream_interpreter
        is_async = inspect.iscoroutinefunction(self.stream_interpreter)

        if is_async:
            log.info(f"Stream interpreter is async: {observed_stream_interpreter}")

        prep = self.prep_vac(request, vector_name)
        log.info(f"Processing prep: {prep}")
        trace = prep["trace"]
        span = prep["span"]
        vac_config = prep["vac_config"]
        all_input = prep["all_input"]

        log.info(f'Streaming data with: {all_input}')
        if span:
            span.update(
                name="start_streaming_chat",
                metadata=vac_config.configs_by_kind,
                input=all_input
            )

        def generate_response_content():
            try:
                if is_async:
                    from queue import Queue, Empty
                    result_queue = Queue()
                    import threading

                    def run_async():
                        async def process_async():
                            try:
                                async_gen = start_streaming_chat_async(
                                    question=all_input["user_input"],
                                    vector_name=vector_name,
                                    qna_func_async=observed_stream_interpreter,
                                    chat_history=all_input["chat_history"],
                                    wait_time=all_input["stream_wait_time"],
                                    timeout=all_input["stream_timeout"],
                                    trace_id=trace.id if trace else None,
                                    **all_input["kwargs"]
                                )
                                log.info(f"{async_gen=}")
                                async for chunk in async_gen:
                                    if isinstance(chunk, dict) and 'answer' in chunk:
                                        if trace:
                                            chunk["trace_id"] = trace.id
                                            chunk["trace_url"] = trace.get_trace_url()
                                            span.end(output=json.dumps(chunk))
                                            trace.update(output=json.dumps(chunk))
                                        archive_qa(chunk, vector_name)
                                        result_queue.put(json.dumps(chunk))
                                    else:
                                        result_queue.put(chunk)
                            except Exception as e:
                                result_queue.put(f"Streaming Error: {str(e)} {traceback.format_exc()}")
                            finally:
                                result_queue.put(None)  # Sentinel
                        asyncio.run(process_async())

                    thread = threading.Thread(target=run_async)
                    thread.start()

                    # Read from the queue and yield results
                    while True:
                        chunk = result_queue.get()
                        if chunk is None:
                            break
                        yield chunk

                    thread.join()
                else:
                    log.info("sync streaming response")
                    for chunk in start_streaming_chat(
                        question=all_input["user_input"],
                        vector_name=vector_name,
                        qna_func=observed_stream_interpreter,
                        chat_history=all_input["chat_history"],
                        wait_time=all_input["stream_wait_time"],
                        timeout=all_input["stream_timeout"],
                        trace_id=trace.id if trace else None,
                        **all_input["kwargs"]
                    ):
                        if isinstance(chunk, dict) and 'answer' in chunk:
                            if trace:
                                chunk["trace_id"] = trace.id
                                chunk["trace_url"] = trace.get_trace_url()
                            archive_qa(chunk, vector_name)
                            if trace:
                                span.end(output=json.dumps(chunk))
                                trace.update(output=json.dumps(chunk))
                            yield json.dumps(chunk)
                        else:
                            yield chunk

            except Exception as e:
                yield f"Streaming Error: {str(e)} {traceback.format_exc()}"

        # Here, the generator function will handle streaming the content to the client.
        response = Response(generate_response_content(), content_type='text/plain; charset=utf-8')
        response.headers['Transfer-Encoding'] = 'chunked'

        log.debug(f"streaming response: {response}")
        if trace:
            span.end(output=response)
            trace.update(output=response)
            self.langfuse_eval_response(trace_id=trace.id, eval_percent=all_input.get('eval_percent'))

        return response

    async def handle_stream_vac_async(self, vector_name):
        observed_stream_interpreter = self.stream_interpreter
        is_async = inspect.iscoroutinefunction(self.stream_interpreter)

        if not is_async:
            raise ValueError(f"Stream interpreter must be async: {observed_stream_interpreter}")

        # Use the async version of prep_vac
        prep = await self.prep_vac_async(request, vector_name)
        log.info(f"Processing async prep: {prep}")
        all_input = prep["all_input"]

        log.info(f'Streaming async data with: {all_input}')

        async def generate_response_content():
            try:
                # Direct async handling without the queue/thread approach
                async_gen = start_streaming_chat_async(
                    question=all_input["user_input"],
                    vector_name=vector_name,
                    qna_func_async=observed_stream_interpreter,
                    chat_history=all_input["chat_history"],
                    wait_time=all_input["stream_wait_time"],
                    timeout=all_input["stream_timeout"],
                    **all_input["kwargs"]
                )
                
                log.info(f"{async_gen=}")
                async for chunk in async_gen:
                    if isinstance(chunk, dict) and 'answer' in chunk:
                        await archive_qa(chunk, vector_name)
                        yield json.dumps(chunk)
                    else:
                        yield chunk

            except Exception as e:
                yield f"Streaming async Error: {str(e)} {traceback.format_exc()}"

        response = Response(generate_response_content(), content_type='text/plain; charset=utf-8')
        response.headers['Transfer-Encoding'] = 'chunked'

        log.debug(f"streaming async response: {response}")

        return response

    @staticmethod
    async def _async_generator_to_stream(async_gen_func):
        """Helper function to stream the async generator's values to the client."""
        async for item in async_gen_func():
            yield item
    
    def langfuse_eval_response(self, trace_id, eval_percent=0.01):
        """
        Sends an evaluation message based on a probability defined by eval_percent.
        
        Args:
            eval_percent (float): The probability (0 to 1) of triggering the evaluation.
            trace_id (str): The trace identifier for the evaluation.
        
        Returns:
            None
        """
        if eval_percent > 1 or eval_percent < 0:
            raise ValueError("eval_percent must be a float between 0 and 1.")

        # Generate a random float between 0 and 1
        random_value = random.random()
        log.info(f"Eval: {trace_id=} {eval_percent=} / {random_value=}")
        # Check if evaluation should be triggered
        if random_value < eval_percent:
            if PubSubManager:
                try:
                    log.info(f"Publishing for eval {trace_id=}")
                    pubsub_manager = PubSubManager("langfuse_evals", pubsub_topic="topicid-to-langfuse-eval")
                    the_data = {"trace_id": trace_id}
                    pubsub_manager.publish_message(the_data)
                except Exception as e:
                    log.warning(f"Could not publish message for 'langfuse_evals' to topicid-to-langfuse-eval - {str(e)}")
        else:
            log.info(f"Did not do Langfuse eval due to random sampling not passed: {eval_percent=}")

    def handle_process_vac(self, vector_name):
        #TODO: handle async
        observed_vac_interpreter = self.vac_interpreter
        prep = self.prep_vac(request, vector_name)
        log.debug(f"Processing prep: {prep}")
        trace = prep["trace"]
        span = prep["span"]
        vac_config: ConfigManager = prep["vac_config"]
        all_input = prep["all_input"]

        try:
            if span:
                gen = span.generation(
                    name="vac_interpreter",
                    metadata=vac_config.configs_by_kind,
                    input = all_input,
                    model=vac_config.vacConfig("model") or vac_config.vacConfig("llm")
                )
            bot_output = observed_vac_interpreter(
                question=all_input["user_input"],
                vector_name=vector_name,
                chat_history=all_input["chat_history"],
                trace_id=trace.id if trace else None,
                **all_input["kwargs"]
            )
            if span:
                gen.end(output=bot_output)
            # {"answer": "The answer", "source_documents": [{"page_content": "The page content", "metadata": "The metadata"}]}
            bot_output = parse_output(bot_output)
            if trace:
                bot_output["trace_id"] = trace.id
                bot_output["trace_url"] = trace.get_trace_url()
            archive_qa(bot_output, vector_name)
            log.info(f'==LLM Q:{all_input["user_input"]} - A:{bot_output}')


        except Exception as err: 
            bot_output = {'answer': f'QNA_ERROR: An error occurred while processing /vac/{vector_name}: {str(err)} traceback: {traceback.format_exc()}'}
            if span:
                gen.end(output=bot_output)

        if trace:
            span.end(output=jsonify(bot_output))
            trace.update(output=jsonify(bot_output)) 
            self.langfuse_eval_response(trace_id=trace.id, eval_percent=all_input.get('eval_percent'))

        # {'answer': 'output'}
        return jsonify(bot_output)

    def check_authentication(self):
        if request.path.startswith('/openai/'):
            log.debug(f'Request headers: {request.headers}')
            # the header forwarded
            auth_header = request.headers.get('X-Forwarded-Authorization')
            if auth_header:
                
                if auth_header.startswith('Bearer '):
                    api_key = auth_header.split(' ')[1]  # Assuming "Bearer <api_key>"
                else:
                    return jsonify({'error': 'Invalid authorization header does not start with "Bearer " - got: {auth_header}'}), 401
                
                endpoints_host = os.getenv('_ENDPOINTS_HOST')
                if not endpoints_host:
                    return jsonify({'error': '_ENDPOINTS_HOST environment variable not found'}), 401
                
                # Check cache first
                current_time = datetime.datetime.now()
                if api_key in api_key_cache:
                    cached_result, cache_time = api_key_cache[api_key]
                    if current_time - cache_time < cache_duration:
                        if not cached_result:
                            return jsonify({'error': 'Invalid cached API key'}), 401
                        else:
                            return  # Valid API key, continue to the endpoint
                    else:
                        # Cache expired, remove from cache
                        del api_key_cache[api_key]
                
                # Validate API key
                is_valid = validate_api_key(api_key, endpoints_host)
                # Update cache
                api_key_cache[api_key] = (is_valid, current_time)
                
                if not is_valid:
                    return jsonify({'error': 'Invalid API key'}), 401
            else:
                return jsonify({'error': 'Missing Authorization header'}), 401

    def openai_health_endpoint():
            return jsonify({'message': 'Success'})

    def handle_openai_compatible_endpoint(self, vector_name=None):
        data = request.get_json()
        log.info(f'openai_compatible_endpoint got data: {data} for vector: {vector_name}')

        vector_name = vector_name or data.pop('model', None)
        messages = data.pop('messages', None)
        chat_history = data.pop('chat_history', None)
        stream = data.pop('stream', False)

        if not messages:
            return jsonify({"error": "No messages provided"}), 400

        user_message = None
        image_uri = None
        mime_type = None

        for msg in reversed(messages):
            if msg['role'] == 'user':
                if isinstance(msg['content'], list):
                    for content_item in msg['content']:
                        if content_item['type'] == 'text':
                            user_message = content_item['text']
                        elif content_item['type'] == 'image_url':
                            base64_data = content_item['image_url']['url']
                            image_uri, mime_type = handle_base64_image(base64_data, vector_name)
                else:
                    user_message = msg['content']
                break

        if not user_message:
            return jsonify({"error": "No user message provided"}), 400
        else:
            log.info(f"User message: {user_message}")
    
        if image_uri:
            data["image_uri"] = image_uri
            data["mime"] = mime_type

        all_input = {
            "user_input": user_message,
            "chat_history": chat_history,
            "kwargs": data
        }

        observed_stream_interpreter = self.stream_interpreter

        response_id = str(uuid.uuid4())
        
        def generate_response_content():
            for chunk in start_streaming_chat(question=user_message,
                                            vector_name=vector_name,
                                            qna_func=observed_stream_interpreter,
                                            chat_history=all_input["chat_history"],
                                            wait_time=all_input.get("stream_wait_time", 1),
                                            timeout=all_input.get("stream_timeout", 60),
                                            trace_id=all_input.get("trace_id"),
                                            **all_input["kwargs"]
                                            ):
                if isinstance(chunk, dict) and 'answer' in chunk:
                    openai_chunk = {
                        "id": response_id,
                        "object": "chat.completion.chunk",
                        "created": str(int(datetime.datetime.now().timestamp())),
                        "model": vector_name,
                        "system_fingerprint": sunholo_version(),
                        "choices": [{
                            "index": 0,
                            "delta": {"content": chunk['answer']},
                            "logprobs": None,
                            "finish_reason": None
                        }]
                    }
                    yield json.dumps(openai_chunk) + "\n"
                else:
                    log.info(f"Unknown chunk: {chunk}")

            final_chunk = {
                "id": response_id,
                "object": "chat.completion.chunk",
                "created": str(int(datetime.datetime.now().timestamp())),
                "model": vector_name,
                "system_fingerprint": sunholo_version(),
                "choices": [{
                    "index": 0,
                    "delta": {},
                    "logprobs": None,
                    "finish_reason": "stop"
                }]
            }
            yield json.dumps(final_chunk) + "\n"

        if stream:
            log.info("Streaming openai chunks")
            return Response(generate_response_content(), content_type='text/plain; charset=utf-8')

        try:
            observed_vac_interpreter = self.vac_interpreter
            bot_output = observed_vac_interpreter(
                question=user_message,
                vector_name=vector_name,
                chat_history=all_input["chat_history"],
                **all_input["kwargs"]
            )
            bot_output = parse_output(bot_output)

            log.info(f"Bot output: {bot_output}")
            if bot_output:
                return self.make_openai_response(user_message, vector_name, bot_output.get('answer', ''))
            else:
                return self.make_openai_response(user_message, vector_name, 'ERROR: could not find an answer')

        except Exception as err:
            log.error(f"OpenAI response error: {str(err)} traceback: {traceback.format_exc()}")
        
            return self.make_openai_response(user_message, vector_name, f'ERROR: {str(err)}')


    def create_langfuse_trace(self, request, vector_name, trace_id):
        try:
            from langfuse import Langfuse
            langfuse = Langfuse()
        except ImportError as err:
            print(f"No langfuse installed for agents.flask.register_qna_routes, install via `pip install sunholo[http]` - {str(err)}")
            
            return None

        user_id = request.headers.get("X-User-ID")
        session_id = request.headers.get("X-Session-ID")
        message_source = request.headers.get("X-Message-Source")

        package_version = sunholo_version()
        tags = [package_version, "autogenerated"]
        if message_source:
            tags.append(message_source)

        return langfuse.trace(
            id=trace_id, 
            name = f"/vac/{vector_name}",
            user_id = user_id,
            session_id = session_id,
            tags = tags,
            release = package_version
        )

    def prep_vac(self, request, vector_name):

        if request.content_type.startswith('application/json'):
            data = request.get_json()
        elif request.content_type.startswith('multipart/form-data'):
            data = request.form.to_dict()
            if 'file' in request.files:
                file = request.files['file']
                if file.filename != '':
                    log.info(f"Found file: {file.filename} to upload to GCS")
                    try:
                        image_uri, mime_type = self.handle_file_upload(file, vector_name)
                        data["image_uri"] = image_uri
                        data["mime"] = mime_type
                    except Exception as e:
                        log.error(traceback.format_exc())
                        return jsonify({'error': str(e), 'traceback': traceback.format_exc()}), 500
                else:
                    log.error("No file selected")
                    return jsonify({"error": "No file selected"}), 400
        else:
            return jsonify({"error": "Unsupported content type"}), 400

        log.info(f"vac/{vector_name} got data: {data}")

        trace = None
        span = None
        if self.add_langfuse_eval:
            trace_id = data.get('trace_id')
            trace = self.create_langfuse_trace(request, vector_name, trace_id)
            log.info(f"Using existing langfuse trace: {trace_id}")
        
        #config, _ = load_config("config/llm_config.yaml")
        try:
            vac_config = ConfigManager(vector_name)
        except Exception as e:
            raise ValueError(f"Unable to find vac_config for {vector_name} - {str(e)}")

        if trace:
            this_vac_config = vac_config.configs_by_kind.get("vacConfig")
            metadata_config=None
            if this_vac_config:
                metadata_config = this_vac_config.get(vector_name)

            trace.update(input=data, metadata=metadata_config)

        user_input = data.pop('user_input').strip()
        stream_wait_time = data.pop('stream_wait_time', 7)
        stream_timeout = data.pop('stream_timeout', 120)
        chat_history = data.pop('chat_history', None)
        eval_percent = data.pop('eval_percent', 0.01)
        vector_name = data.pop('vector_name', vector_name)
        data.pop('trace_id', None) # to ensure not in kwargs

        paired_messages = extract_chat_history_with_cache(chat_history)

        all_input = {'user_input': user_input, 
                     'vector_name': vector_name, 
                     'chat_history': paired_messages, 
                     'stream_wait_time': stream_wait_time,
                     'stream_timeout': stream_timeout,
                     'eval_percent': eval_percent,
                     'kwargs': data}

        if trace:
            span = trace.span(
                name="VAC",
                metadata=vac_config.configs_by_kind,
                input = all_input
            )
    
        return {
            "trace": trace,
            "span": span,
            "all_input": all_input,
            "vac_config": vac_config
        }

    async def prep_vac_async(self, request, vector_name):
        """Async version of prep_vac."""
        # Parse request data
        if request.content_type.startswith('application/json'):
            data = request.get_json()
        elif request.content_type.startswith('multipart/form-data'):
            data = request.form.to_dict()
            if 'file' in request.files:
                file = request.files['file']
                if file.filename != '':
                    log.info(f"Found file: {file.filename} to upload to GCS")
                    try:
                        # Make file upload async if possible
                        image_uri, mime_type = await self.handle_file_upload_async(file, vector_name)
                        data["image_uri"] = image_uri
                        data["mime"] = mime_type
                    except Exception as e:
                        log.error(traceback.format_exc())
                        return jsonify({'error': str(e), 'traceback': traceback.format_exc()}), 500
                else:
                    log.error("No file selected")
                    return jsonify({"error": "No file selected"}), 400
        else:
            return jsonify({"error": "Unsupported content type"}), 400

        log.info(f"vac/{vector_name} got data: {data}")

        # Run these operations concurrently
        tasks = []
        
        # Extract other data while configs load
        user_input = data.pop('user_input').strip()
        stream_wait_time = data.pop('stream_wait_time', 7)
        stream_timeout = data.pop('stream_timeout', 120)
        chat_history = data.pop('chat_history', None)
        vector_name_param = data.pop('vector_name', vector_name)
        data.pop('trace_id', None)  # to ensure not in kwargs
        
        # Task 3: Process chat history
        chat_history_task = asyncio.create_task(extract_chat_history_async_cached(chat_history))
        tasks.append(chat_history_task)
        
        # Await all tasks concurrently
        results = await asyncio.gather(*tasks, return_exceptions=True)

        paired_messages = results[0] if not isinstance(results[0], Exception) else []
        
        # Only create span after we have trace
        all_input = {
            'user_input': user_input, 
            'vector_name': vector_name_param, 
            'chat_history': paired_messages, 
            'stream_wait_time': stream_wait_time,
            'stream_timeout': stream_timeout,
            'kwargs': data
        }
        
        return {
            "all_input": all_input
        }

    def handle_file_upload(self, file, vector_name):
        try:
            file.save(file.filename)
            image_uri = add_file_to_gcs(file.filename, vector_name)
            os.remove(file.filename)  # Clean up the saved file
            return image_uri, file.mimetype
        except Exception as e:
            raise Exception(f'File upload failed: {str(e)}')

    async def _initialize_mcp_servers(self):
        """Initialize connections to configured MCP servers."""
        for server_config in self.mcp_servers:
            try:
                await self.mcp_client_manager.connect_to_server(
                    server_name=server_config["name"],
                    command=server_config["command"],
                    args=server_config.get("args", [])
                )
                log.info(f"Connected to MCP server: {server_config['name']}")
            except Exception as e:
                log.error(f"Failed to connect to MCP server {server_config['name']}: {e}")

    
    def handle_mcp_list_tools(self, server_name: Optional[str] = None):
        """List available MCP tools."""
        async def get_tools():
            tools = await self.mcp_client_manager.list_tools(server_name)
            return [
                {
                    "name": tool.name,
                    "description": tool.description,
                    "inputSchema": tool.inputSchema,
                    "server": tool.metadata.get("server") if tool.metadata else server_name
                }
                for tool in tools
            ]
        
        # Run async in sync context
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        try:
            tools = loop.run_until_complete(get_tools())
            return jsonify({"tools": tools})
        finally:
            loop.close()
    
    def handle_mcp_call_tool(self):
        """Call an MCP tool."""
        data = request.get_json()
        server_name = data.get("server")
        tool_name = data.get("tool")
        arguments = data.get("arguments", {})
        
        if not server_name or not tool_name:
            return jsonify({"error": "Missing 'server' or 'tool' parameter"}), 400
        
        async def call_tool():
            try:
                result = await self.mcp_client_manager.call_tool(server_name, tool_name, arguments)
                
                # Convert result to JSON-serializable format
                if hasattr(result, 'content'):
                    # Handle different content types
                    if hasattr(result.content, 'text'):
                        return {"result": result.content.text}
                    elif hasattr(result.content, 'data'):
                        return {"result": result.content.data}
                    else:
                        return {"result": str(result.content)}
                else:
                    return {"result": str(result)}
                    
            except Exception as e:
                return {"error": str(e)}
        
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        try:
            result = loop.run_until_complete(call_tool())
            if "error" in result:
                return jsonify(result), 500
            return jsonify(result)
        finally:
            loop.close()
    
    def handle_mcp_list_resources(self):
        """List available MCP resources."""
        server_name = request.args.get("server")
        
        async def get_resources():
            resources = await self.mcp_client_manager.list_resources(server_name)
            return [
                {
                    "uri": resource.uri,
                    "name": resource.name,
                    "description": resource.description,
                    "mimeType": resource.mimeType,
                    "server": resource.metadata.get("server") if resource.metadata else server_name
                }
                for resource in resources
            ]
        
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        try:
            resources = loop.run_until_complete(get_resources())
            return jsonify({"resources": resources})
        finally:
            loop.close()
    
    def handle_mcp_read_resource(self):
        """Read an MCP resource."""
        data = request.get_json()
        server_name = data.get("server")
        uri = data.get("uri")
        
        if not server_name or not uri:
            return jsonify({"error": "Missing 'server' or 'uri' parameter"}), 400
        
        async def read_resource():
            try:
                contents = await self.mcp_client_manager.read_resource(server_name, uri)
                return {
                    "contents": [
                        {"text": content.text} if hasattr(content, 'text') else {"data": str(content)}
                        for content in contents
                    ]
                }
            except Exception as e:
                return {"error": str(e)}
        
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        try:
            result = loop.run_until_complete(read_resource())
            if "error" in result:
                return jsonify(result), 500
            return jsonify(result)
        finally:
            loop.close()
    
    def handle_mcp_server(self):
        """Handle MCP server requests using HTTP transport."""
        if not self.vac_mcp_server:
            return jsonify({"error": "MCP server not enabled"}), 501
        
        import json as json_module
        
        # Handle streaming for HTTP transport
        if request.method == 'POST':
            try:
                # Get the JSON-RPC request
                data = request.get_json()
                log.info(f"MCP server received: {data}")
                
                # Create an async handler for the request
                async def process_request():
                    # Handle JSON-RPC requests directly
                    if not isinstance(data, dict) or "method" not in data:
                        raise ValueError("Invalid JSON-RPC request")
                    
                    method = data["method"]
                    params = data.get("params", {})
                    request_id = data.get("id")
                    
                    try:
                        # Handle different MCP methods
                        if method == "initialize":
                            # Handle MCP protocol initialization
                            protocol_version = params.get("protocolVersion", "2025-06-18")
                            capabilities = params.get("capabilities", {})
                            client_info = params.get("clientInfo", {})
                            
                            log.info(f"MCP client initializing: {client_info} with protocol version: {protocol_version}")
                            
                            return {
                                "jsonrpc": "2.0",
                                "result": {
                                    "protocolVersion": "2025-06-18",
                                    "capabilities": {
                                        "tools": {}
                                    },
                                    "serverInfo": {
                                        "name": "sunholo-vac-server",
                                        "version": sunholo_version()
                                    }
                                },
                                "id": request_id
                            }
                            
                        elif method == "tools/list":
                            # Create mock tools list based on VACMCPServer capabilities
                            tools = [
                                {
                                    "name": "vac_stream",
                                    "description": "Stream responses from a Sunholo VAC (Virtual Agent Computer)",
                                    "inputSchema": {
                                        "type": "object",
                                        "properties": {
                                            "vector_name": {"type": "string", "description": "Name of the VAC to interact with"},
                                            "user_input": {"type": "string", "description": "The user's question or input"},
                                            "chat_history": {"type": "array", "description": "Previous conversation history", "default": []},
                                            "stream_wait_time": {"type": "number", "description": "Time to wait between stream chunks", "default": 7},
                                            "stream_timeout": {"type": "number", "description": "Maximum time to wait for response", "default": 120}
                                        },
                                        "required": ["vector_name", "user_input"]
                                    }
                                }
                            ]
                            
                            # Add vac_query tool if interpreter is available
                            if self.vac_mcp_server.vac_interpreter:
                                tools.append({
                                    "name": "vac_query",
                                    "description": "Query a Sunholo VAC (non-streaming)",
                                    "inputSchema": {
                                        "type": "object",
                                        "properties": {
                                            "vector_name": {"type": "string", "description": "Name of the VAC to interact with"},
                                            "user_input": {"type": "string", "description": "The user's question or input"},
                                            "chat_history": {"type": "array", "description": "Previous conversation history", "default": []}
                                        },
                                        "required": ["vector_name", "user_input"]
                                    }
                                })
                            
                            return {
                                "jsonrpc": "2.0",
                                "result": {"tools": tools},
                                "id": request_id
                            }
                            
                        elif method == "tools/call":
                            # Handle tool calls
                            tool_name = params.get("name")
                            arguments = params.get("arguments", {})
                            
                            if not tool_name:
                                raise ValueError("Missing tool name")
                            
                            # Call the appropriate handler method
                            if tool_name == "vac_stream":
                                result = await self.vac_mcp_server._handle_vac_stream(arguments)
                            elif tool_name == "vac_query" and self.vac_mcp_server.vac_interpreter:
                                result = await self.vac_mcp_server._handle_vac_query(arguments)
                            else:
                                raise ValueError(f"Unknown tool: {tool_name}")
                            
                            return {
                                "jsonrpc": "2.0", 
                                "result": {"content": [item.model_dump() for item in result]},
                                "id": request_id
                            }
                            
                        else:
                            raise ValueError(f"Unknown method: {method}")
                            
                    except Exception as e:
                        log.error(f"Error handling MCP method {method}: {e}")
                        return {
                            "jsonrpc": "2.0",
                            "error": {
                                "code": -32603,
                                "message": str(e)
                            },
                            "id": request_id
                        }
                
                # Run the async handler
                loop = asyncio.new_event_loop()
                asyncio.set_event_loop(loop)
                try:
                    response_data = loop.run_until_complete(process_request())
                    return jsonify(response_data)

                except Exception as e:
                    log.error(f"MCP server error: {str(e)}")
                    return jsonify({
                        "jsonrpc": "2.0",
                        "error": {
                            "code": -32603,
                            "message": f"Internal error: {str(e)}"
                        },
                        "id": data.get("id") if isinstance(data, dict) else None
                    }), 500
                finally:
                    loop.close()
                        
            except Exception as e:
                log.error(f"MCP server error: {str(e)}")
                return jsonify({
                    "jsonrpc": "2.0",
                    "error": {
                        "code": -32603,
                        "message": f"Internal error: {str(e)}"
                    },
                    "id": data.get("id") if isinstance(data, dict) else None
                }), 500
                    
        else:
            # GET request - return server information
            return jsonify({
                "name": "sunholo-vac-server",
                "version": "1.0.0",
                "transport": "http",
                "endpoint": "/mcp",
                "tools": ["vac_stream", "vac_query"] if self.vac_interpreter else ["vac_stream"]
            })
    
    def _get_or_create_a2a_agent(self):
        """Get or create the A2A agent instance with current request context."""
        if not self.enable_a2a_agent or not VACA2AAgent:
            return None
        
        if self.vac_a2a_agent is None:
            # Extract base URL from current request
            base_url = request.url_root.rstrip('/')
            
            self.vac_a2a_agent = VACA2AAgent(
                base_url=base_url,
                stream_interpreter=self.stream_interpreter,
                vac_interpreter=self.vac_interpreter,
                vac_names=self.a2a_vac_names
            )
        
        return self.vac_a2a_agent
    
    def handle_a2a_agent_card(self):
        """Handle A2A agent card discovery request."""
        agent = self._get_or_create_a2a_agent()
        if not agent:
            return jsonify({"error": "A2A agent not enabled"}), 501
        
        return jsonify(agent.get_agent_card())
    
    def handle_a2a_task_send(self):
        """Handle A2A task send request."""
        agent = self._get_or_create_a2a_agent()
        if not agent:
            return jsonify({"error": "A2A agent not enabled"}), 501
        
        try:
            data = request.get_json()
            if not data:
                return jsonify({
                    "jsonrpc": "2.0",
                    "error": {
                        "code": -32700,
                        "message": "Parse error: Invalid JSON"
                    },
                    "id": None
                }), 400
            
            # Run async handler
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            try:
                response = loop.run_until_complete(agent.handle_task_send(data))
                return jsonify(response)
            finally:
                loop.close()
                
        except Exception as e:
            log.error(f"A2A task send error: {e}")
            return jsonify({
                "jsonrpc": "2.0",
                "error": {
                    "code": -32603,
                    "message": f"Internal error: {str(e)}"
                },
                "id": data.get("id") if 'data' in locals() else None
            }), 500
    
    def handle_a2a_task_send_subscribe(self):
        """Handle A2A task send with subscription (SSE)."""
        agent = self._get_or_create_a2a_agent()
        if not agent:
            return jsonify({"error": "A2A agent not enabled"}), 501
        
        try:
            data = request.get_json()
            if not data:
                def error_generator():
                    yield "data: {\"error\": \"Parse error: Invalid JSON\"}\n\n"
                
                return Response(error_generator(), content_type='text/event-stream')
            
            # Create async generator for SSE
            async def sse_generator():
                async for chunk in agent.handle_task_send_subscribe(data):
                    yield chunk
            
            def sync_generator():
                loop = asyncio.new_event_loop()
                asyncio.set_event_loop(loop)
                try:
                    async_gen = sse_generator()
                    while True:
                        try:
                            chunk = loop.run_until_complete(async_gen.__anext__())
                            yield chunk
                        except StopAsyncIteration:
                            break
                finally:
                    loop.close()
            
            return Response(sync_generator(), content_type='text/event-stream')
            
        except Exception as e:
            log.error(f"A2A task send subscribe error: {e}")
            def error_generator(err):
                yield f"data: {{\"error\": \"Internal error: {str(err)}\"}}\n\n"
            
            return Response(error_generator(e), content_type='text/event-stream')
    
    def handle_a2a_task_get(self):
        """Handle A2A task get request."""
        agent = self._get_or_create_a2a_agent()
        if not agent:
            return jsonify({"error": "A2A agent not enabled"}), 501
        
        try:
            data = request.get_json()
            if not data:
                return jsonify({
                    "jsonrpc": "2.0",
                    "error": {
                        "code": -32700,
                        "message": "Parse error: Invalid JSON"
                    },
                    "id": None
                }), 400
            
            # Run async handler
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            try:
                response = loop.run_until_complete(agent.handle_task_get(data))
                return jsonify(response)
            finally:
                loop.close()
                
        except Exception as e:
            log.error(f"A2A task get error: {e}")
            return jsonify({
                "jsonrpc": "2.0",
                "error": {
                    "code": -32603,
                    "message": f"Internal error: {str(e)}"
                },
                "id": data.get("id") if 'data' in locals() else None
            }), 500
    
    def handle_a2a_task_cancel(self):
        """Handle A2A task cancel request."""
        agent = self._get_or_create_a2a_agent()
        if not agent:
            return jsonify({"error": "A2A agent not enabled"}), 501
        
        try:
            data = request.get_json()
            if not data:
                return jsonify({
                    "jsonrpc": "2.0",
                    "error": {
                        "code": -32700,
                        "message": "Parse error: Invalid JSON"
                    },
                    "id": None
                }), 400
            
            # Run async handler
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            try:
                response = loop.run_until_complete(agent.handle_task_cancel(data))
                return jsonify(response)
            finally:
                loop.close()
                
        except Exception as e:
            log.error(f"A2A task cancel error: {e}")
            return jsonify({
                "jsonrpc": "2.0",
                "error": {
                    "code": -32603,
                    "message": f"Internal error: {str(e)}"
                },
                "id": data.get("id") if 'data' in locals() else None
            }), 500
    
    def handle_a2a_push_notification(self):
        """Handle A2A push notification settings."""
        agent = self._get_or_create_a2a_agent()
        if not agent:
            return jsonify({"error": "A2A agent not enabled"}), 501
        
        try:
            data = request.get_json()
            if not data:
                return jsonify({
                    "jsonrpc": "2.0",
                    "error": {
                        "code": -32700,
                        "message": "Parse error: Invalid JSON"
                    },
                    "id": None
                }), 400
            
            # Run async handler
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            try:
                response = loop.run_until_complete(agent.handle_push_notification_set(data))
                return jsonify(response)
            finally:
                loop.close()
                
        except Exception as e:
            log.error(f"A2A push notification error: {e}")
            return jsonify({
                "jsonrpc": "2.0",
                "error": {
                    "code": -32603,
                    "message": f"Internal error: {str(e)}"
                },
                "id": data.get("id") if 'data' in locals() else None
            }), 500


#   Copyright [2024] [Holosun ApS]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""A2A (Agent-to-Agent) protocol integration for Sunholo."""

try:
    from .agent_card import AgentCardGenerator
    from .task_manager import A2ATaskManager
    from .vac_a2a_agent import VACA2AAgent
except (ImportError, SyntaxError) as e:
    # Handle missing dependencies or syntax errors gracefully
    AgentCardGenerator = None
    VACA2AAgent = None
    A2ATaskManager = None

__all__ = ['AgentCardGenerator', 'VACA2AAgent', 'A2ATaskManager']


import os
try:
    from azure.identity import DefaultAzureCredential, ClientSecretCredential
except ImportError:
    DefaultAzureCredential = None
    ClientSecretCredential = None

from ..custom_logging import log

def azure_auth():
    """
    Will attempt to authenticate using default credentials first (e.g. you are running within Azure Container Apps or similar)

    If default credentials are not available, will attempt to authenticate via env vars - set up via:

    ```bash
    az ad sp create-for-rbac --name "myApp" --role contributor \
        --scopes /subscriptions/{subscription-id}/resourceGroups/{resource-group} \
        --sdk-auth

    export AZURE_CLIENT_ID="your-client-id"
    export AZURE_CLIENT_SECRET="your-client-secret"
    export AZURE_TENANT_ID="your-tenant-id"
    ```

    """
    if DefaultAzureCredential is None: 
        raise ImportError("Azure identity credentials library needed - install via `pip install sunholo[azure]`")
    
    # Use DefaultAzureCredential to authenticate
    try:
        credential = DefaultAzureCredential()
        return credential
    
    except Exception as e:
        log.error(f"Failed to authenticate with default credentials: {str(e)}")
        log.info("Attempting to authenticate using ClientSecretCredential")

        # Use ClientSecretCredential to authenticate with a service principal
        client_id = os.getenv("AZURE_CLIENT_ID")
        client_secret = os.getenv("AZURE_CLIENT_SECRET")
        tenant_id = os.getenv("AZURE_TENANT_ID")

        if not client_id or not client_secret or not tenant_id:
            log.error("Service principal credentials are not set in environment variables")
            return None

        if ClientSecretCredential is None:
            raise ImportError("Azure identity credentials library needed - install via `pip install sunholo[azure]`")

        try:
            credential = ClientSecretCredential(
                client_id=client_id,
                client_secret=client_secret,
                tenant_id=tenant_id
            )
            return credential
        except Exception as e:
            log.error(f"Failed to authenticate with service principal: {str(e)}")
            return None
        



import os.path
import argparse
from typing import List, Optional
import sys
from ..custom_logging import log

try:
    import PIL.Image
    from ollama import generate
except ImportError:
    generate = None

CHAT_MODEL_NAME = os.getenv("MODEL_NAME_LATEST")

def chat_ollama(msg, model_name, the_images=None):
    
    if not generate:
        raise ImportError("Import ollama via `pip install ollama`")
    
    chat_images = []
    if the_images:
        for the_image in the_images:
            chat_image = PIL.Image.open(the_image)
            chat_images.append(chat_image)

    log.info(f"Ollama [{model_name}]: Chatting...{msg=}")
    for response in generate(model_name, msg, images=chat_images, stream=True):
        print(response['response'], end='', flush=True)

def main():
    parser = argparse.ArgumentParser(description='Chat with Ollama models from the command line')
    parser.add_argument('--model', '-m', type=str, default=CHAT_MODEL_NAME,
                        help='Model name to use (defaults to MODEL_NAME_LATEST env var)')
    parser.add_argument('--images', '-i', type=str, nargs='+',
                        help='Image file paths to include in the prompt')
    parser.add_argument('--message', '-p', type=str,
                        help='Message to send')
    
    args = parser.parse_args()
    
    if not args.model:
        print("Error: No model specified. Either set MODEL_NAME_LATEST environment variable or use --model flag.")
        sys.exit(1)
    
    # If no message provided via args, read from stdin
    if not args.message:
        print(f"Enter your message to {args.model} (Ctrl+D to send):")
        user_input = sys.stdin.read().strip()
    else:
        user_input = args.message
    
    if not user_input:
        print("Error: Empty message. Exiting.")
        sys.exit(1)
    
    try:
        chat_ollama(user_input, args.model, args.images)
        print()  # Add a newline after the response
    except ImportError as e:
        print(f"Error: {e}")
        sys.exit(1)
    except Exception as e:
        print(f"Error: {e}")
        sys.exit(1)

if __name__ == "__main__":
    # uv run src/sunholo/ollama/ollama_images.py --model=gemma3:12b - chat and then CTRL+D
    # uv run src/sunholo/ollama/ollama_images.py --model gemma3:12b --message "Tell me about quantum computing"


    main()


from ..agents import send_to_qa, handle_special_commands
from ..streaming import generate_proxy_stream, can_agent_stream
from ..utils.user_ids import generate_user_id
from ..utils import ConfigManager
from ..utils.api_key import has_multivac_api_key
from ..custom_logging import log
from ..qna.parsers import parse_output
from ..gcs.add_file import add_file_to_gcs
from .run_proxy import clean_proxy_list, start_proxy, stop_proxy
from ..invoke import invoke_vac
from ..utils.big_context import has_text_extension, merge_text_files, load_gitignore_patterns, build_file_tree
import tempfile

import uuid
import os
import sys
import subprocess
import json
from pathlib import Path

from rich import print
from .sun_rich import console

from rich.prompt import Prompt
from rich.panel import Panel
from rich.text import Text
from rich.table import Table

def read_and_add_to_user_input(user_input):
    read_input = None

    path = user_input.split(" ", 1)[1] if " " in user_input else None
    if not path:
        console.print("[bold red]Please provide a valid file or folder path.[/bold red]")
        return None

    if os.path.isfile(path):
        if not has_text_extension(path):
            console.print("[bold red]Unsupported file type. Please provide a text file or preprocess to text, or use !upload (e.g. images) or `sunholo embed`.[/bold red]")
            return None

        try:
            with open(path, 'r', encoding='utf-8') as file:
                file_content = file.read()
            read_input = file_content
            console.print(f"[bold yellow]File content from {path} read into user_input: [{len(read_input.split())}] words[/bold yellow]")
        except FileNotFoundError:
            console.print("[bold red]File not found. Please check the path and try again.[/bold red]")
            return None
        except IOError:
            console.print("[bold red]File could not be read. Please ensure it is a readable text file.[/bold red]")
            return None
    elif os.path.isdir(path):
        patterns = []
        gitignore_path = os.path.join(path, '.gitignore')
    
        if os.path.exists(gitignore_path):
            patterns = load_gitignore_patterns(gitignore_path)
    
        try:
            with tempfile.NamedTemporaryFile(delete=False, mode='w+', encoding='utf-8') as temp_file:
                temp_file_path = temp_file.name
                file_tree = merge_text_files(path, temp_file_path, patterns)
                console.print(f"[bold yellow]Contents of the folder '{path}' have been merged add added to input.[/bold yellow]")
                console.print("\n".join(file_tree))
                temp_file.seek(0)
                read_input = temp_file.read()
                console.print(f"[bold yellow]Total words: [{len(read_input.split())}] - watch out for high token costs! Use !clear_read to reset[/bold yellow]")
            os.remove(temp_file_path)  # Clean up the temporary file
        except Exception as e:
            console.print(f"[bold red]An error occurred while reading the folder: {str(e)}[/bold red]")
            return None
    else:
        console.print("[bold red]The provided path is neither a file nor a folder. Please check the path and try again.[/bold red]")
        return None
    
    return read_input

def get_service_url(vac_name, project, region, no_config=False):

    if no_config:
        agent_name = vac_name
    else:
        agent_name = ConfigManager(vac_name).vacConfig("agent")

    proxies = clean_proxy_list()
    if agent_name in proxies:
        port = proxies[agent_name]['port']
        url = f"http://127.0.0.1:{port}"
    else:
        if agent_name:
            console.print(f"No proxy found running for service: [bold orange]'{agent_name}'[/bold orange] required for [bold orange]{vac_name}[/bold orange] - attempting to connect")
            url = start_proxy(agent_name, region, project)
        else:
            console.print(f"No config for [bold orange]'{vac_name}'[/bold orange] - can't start proxy")
            sys.exit(1)

    return url

def handle_file_upload(file, vector_name):
    if not Path(file).is_file():
        return None
    
    agent_name = ConfigManager(vector_name).vacConfig("agent")
    # vertex can't handle directories
    bucket_filepath = f"{vector_name}/uploads/{os.path.basename(file)}" if agent_name != "vertex-genai" else os.path.basename(file)

    file_url = add_file_to_gcs(file, 
                               vector_name=vector_name, 
                               metadata={"type": "cli"},
                               bucket_filepath=bucket_filepath)
    
    return file_url

def stream_chat_session(service_url, service_name, stream=True):

    user_id = generate_user_id()
    chat_history = []
    agent_name = ConfigManager(service_name).vacConfig("agent")
    file_reply = None
    read_file = None
    read_file_count = None
    while True:
        session_id = str(uuid.uuid4())
        user_input = Prompt.ask("[bold cyan]You[/bold cyan]")
        if user_input.lower() in ["exit", "quit"]:
            console.print("[bold red]Exiting chat session.[/bold red]")
            break

        special_reply = handle_special_commands(
            user_input, 
            vector_name=service_name,
            chat_history=chat_history)

        if special_reply:
             console.print(f"[bold yellow]{service_name}:[/bold yellow] {special_reply}", end='\n')
             continue

        if user_input.lower().startswith("!read"):
            read_file = read_and_add_to_user_input(user_input)
            if read_file:
                read_file_count = len(read_file.split())
            continue

        if user_input.lower().startswith("!ls"):
            items = os.listdir(os.getcwd())
            for item in items:
                console.print(item)
            continue

        if user_input.lower().startswith("!tree"):
            tree = build_file_tree(os.getcwd(), patterns=[])
            console.print(tree)
            continue

        if user_input.lower().startswith("!upload"):
            file_path = user_input.split(" ", 1)[1] if " " in user_input else None
            if not file_path:
                console.print("[bold red]Please provide a valid file path.[/bold red]")
                continue

            try:
                file_reply = handle_file_upload(file_path, vector_name=service_name)
                if not file_reply:
                    console.print("[bold red]Invalid file upload[/bold red]")
                    continue
                
                console.print(f"[bold yellow]{service_name}:[/bold yellow] Uploaded {file_path} to {file_reply} - image will be sent each reply until you issue '!clear_upload' ", end='\n')
            
            except FileNotFoundError:
                console.print("[bold red]File not found. Please check the path and try again.[/bold red]")

            # file_reply stays for each message from now on
            continue 

        if user_input.lower().startswith("!clear_upload"):
            console.print("[bold yellow]File upload path cleared.[/bold yellow]")
            file_path = None
            continue

        if user_input.lower().startswith("!clear_read"):
            console.print("[bold yellow]Read in file(s) cleared.[/bold yellow]")
            read_file = None
            read_file_count = None
            continue

        if user_input.lower().startswith("!"):
            console.print("[bold red]Could find no valid chat command for you, sorry[/bold red]")
            continue
        
        if read_file:
            user_input = f"<user added file>{read_file}</user added file>\n{user_input}"
        
        # guardrail
        if len(user_input)> 1000000:
            console.print("[bold red]Over 1 million characters in user_input, aborting as probably unintentional. Use API directly instead.[/bold red]")
            continue

        if not stream:
            vac_response = send_to_qa(user_input,
                vector_name=service_name,
                chat_history=chat_history,
                message_author=user_id,
                #TODO: populate these
                image_url=file_reply,
                source_filters=None,
                search_kwargs=None,
                private_docs=None,
                whole_document=False,
                source_filters_and_or=False,
                # system kwargs
                configurable={
                    "vector_name": service_name,
                },
                user_id=user_id,
                session_id=session_id, 
                message_source="cli",
                override_endpoint=service_url)
            
            # ensures {'answer': answer}
            answer = parse_output(vac_response)
            
            console.print(f"[bold yellow]{service_name}:[/bold yellow] {answer.get('answer')}", end='\n')
        else:

            def stream_response():
                generate = generate_proxy_stream(
                    send_to_qa,
                    user_input,
                    vector_name=service_name,
                    chat_history=chat_history,
                    generate_f_output=lambda x: x,  # Replace with actual processing function
                    stream_wait_time=0.5,
                    stream_timeout=120 if agent_name != "vertex-genai" else 1200,
                    message_author=user_id,
                    #TODO: populate these
                    image_url=file_reply,
                    source_filters=None,
                    search_kwargs=None,
                    private_docs=None,
                    whole_document=False,
                    source_filters_and_or=False,
                    # system kwargs
                    configurable={
                        "vector_name": service_name,
                    },
                    user_id=user_id,
                    session_id=session_id, 
                    message_source="cli",
                    override_endpoint=service_url
                )
                for part in generate():
                    yield part

            response_started = False
            vac_response = ""

            
            thinking = "[bold orange]Thinking...[/bold orange]"
            if file_reply:
                thinking = f"[bold orange]Thinking with upload {file_reply} - issue !clear_upload to remove...[/bold orange]"
            
            if read_file:
                thinking = f"{thinking} - [bold orange]additional [{read_file_count}] words added via !read_file contents - issue !clear_read to remove[/bold orange]"

            with console.status(thinking, spinner="star") as status:
                for token in stream_response():
                    if not response_started:
                        status.stop()
                        console.print(f"[bold yellow]{service_name}:[/bold yellow] ", end='')
                        response_started = True

                    if isinstance(token, bytes):
                        token = token.decode('utf-8')
                    console.print(token, end='')
                    vac_response += token

            response_started = False

        chat_history.append({"name": "Human", "content": user_input})
        chat_history.append({"name": "AI", "content": vac_response})
        
        console.print()
        console.rule()

def headless_mode(service_url, service_name, user_input, chat_history=None, stream=True):
    chat_history = chat_history or []

    user_id = generate_user_id()
    session_id = str(uuid.uuid4())

    if not stream:
        vac_response = send_to_qa(user_input,
            vector_name=service_name,
            chat_history=chat_history,
            message_author=user_id,
            #TODO: populate these
            image_url=None,
            source_filters=None,
            search_kwargs=None,
            private_docs=None,
            whole_document=False,
            source_filters_and_or=False,
            # system kwargs
            configurable={
                "vector_name": service_name,
            },
            user_id=user_id,
            session_id=session_id, 
            message_source="cli",
            override_endpoint=service_url)
        
        # ensures {'answer': answer}
        answer = parse_output(vac_response)
        
        console.print(answer.get('answer'))
    else:
        def stream_response():
            generate = generate_proxy_stream(
                    send_to_qa,
                    user_input,
                    vector_name=service_name,
                    chat_history=chat_history,
                    generate_f_output=lambda x: x,  # Replace with actual processing function
                    stream_wait_time=0.5,
                    stream_timeout=120,
                    message_author=user_id,
                    #TODO: populate these
                    image_url=None,
                    source_filters=None,
                    search_kwargs=None,
                    private_docs=None,
                    whole_document=False,
                    source_filters_and_or=False,
                    # system kwargs
                    configurable={
                        "vector_name": service_name,
                    },
                    user_id=user_id,
                    session_id=session_id, 
                    message_source="cli",
                    override_endpoint=service_url
            )
            for part in generate():
                yield part

        answer = ""

        for token in stream_response():
            if isinstance(token, bytes):
                token = token.decode('utf-8')
            print(token, end='', flush=True)
            if isinstance(token, dict):
                # ?
                pass
            elif isinstance(token, str):
                answer += token

    if answer:
        chat_history.append({"name": "Human", "content": user_input})
        chat_history.append({"name": "AI", "content": answer})
    print()  # For new line after streaming ends

    return chat_history

def resolve_service_url(args, no_config=False):
    """
    no_config: some VACs do not have an entry in the config file e.g. chunker, embedder etc.
    """
    if args.url_override:

        return args.url_override
    
    config = ConfigManager(args.vac_name)
    global_config = ConfigManager("global")

    agent_name = config.vacConfig("agent")
    agent_url = config.vacConfig("agent_url")

    if agent_url:
        console.print("Found agent_url within vacConfig: {agent_url}")
    
    # via public cloud endpoints - assumes no gcloud auth
    if has_multivac_api_key():
        log.debug("Found MULTIVAC_API_KEY")
        gcp_config = global_config.vacConfig("gcp_config")
        endpoints_base_url = gcp_config.get("endpoints_base_url")
        if not endpoints_base_url:
            console.print("[bold red]MULTIVAC_API_KEY env var is set but no config.gcp_config.endpoints_base_url can be found[/bold red]")
            sys.exit(1)

        return f"{endpoints_base_url}/v1/{agent_name}"

    # via direct access to agent url - requires gcloud auth access
    elif args.no_proxy:
        
        service_url = agent_url or get_cloud_run_service_url(args.project, args.region, agent_name)
        console.print(f"No proxy, connecting directly to {service_url}")

        return service_url
    
    else:
        # via gcloud proxy - requires gcloud auth access
        try:
            service_url = get_service_url(args.vac_name, args.project, args.region, no_config=no_config)
        except ValueError as e:
            console.print(f"[bold red]ERROR: Could not start {args.vac_name} proxy URL: {str(e)}[/bold red]")
            sys.exit(1)
    
        return service_url

def vac_command(args):

    

    if args.action == 'list':

        list_cloud_run_services(args.project, args.region)

        return
    
    elif args.action == 'get-url':
        service_url = resolve_service_url(args)
        console.print(service_url)

        return
    
    elif args.action == 'chat':
        config = ConfigManager(args.vac_name)
        service_url = resolve_service_url(args)
        agent_name   = config.vacConfig("agent")

        streamer = can_agent_stream(agent_name)
        log.debug(f"streamer: {streamer}")
        if not streamer:
            console.print(f"Non streaming agent: {args.vac_name}")

        if args.headless:
            headless_mode(service_url, args.vac_name, args.user_input, args.chat_history, stream=streamer)
        else:
            display_name = config.vacConfig("display_name")
            description  = config.vacConfig("description")
            endpoints_config = config.agentConfig(agent_name)

            post_endpoints = endpoints_config['post']

            display_endpoints = ' '.join(f"{key}: {value}" for key, value in post_endpoints.items())
            display_endpoints = display_endpoints.replace("{stem}", service_url).replace("{vector_name}", args.vac_name)

            if agent_name == "langserve":
                subtitle = f"{service_url}/{args.vac_name}/playground/"
            else:
                subtitle = display_endpoints

            print(
                Panel(description or "Starting VAC chat session", 
                    title=display_name or args.vac_name,
                    subtitle=subtitle)
                    )

            stream_chat_session(service_url, args.vac_name, stream=streamer)
        
        stop_proxy(agent_name, stop_local=False)

    elif args.action == 'invoke':
        service_url = resolve_service_url(args, no_config=True)

        invoke_vac(service_url, args.data, is_file=args.is_file)

def list_cloud_run_services(project, region):
    """
    Lists all Cloud Run services the user has access to in a specific project and region.

    Args:
        project (str): The GCP project ID.
        region (str): The region of the Cloud Run services.
    """

        # point or star?
    with console.status("[bold orange]Listing Cloud Run Services[/bold orange]", spinner="star") as status:
        try:
            result = subprocess.run(
                ["gcloud", "run", "services", "list", "--project", project, "--region", region, "--format=json"],
                stdout=subprocess.PIPE, stderr=subprocess.PIPE, timeout=30
            )
            if result.returncode != 0:
                status.stop()
                console.print(f"[bold red]ERROR: Unable to list Cloud Run services: {result.stderr.decode()}[/bold red]")
                return

            services = json.loads(result.stdout.decode())
            if not services:
                status.stop()
                console.print("[bold red]No Cloud Run services found.[/bold red]")
                return

            proxies = clean_proxy_list()
            status.stop()

            table = Table(title="VAC Cloud Run Services")
            table.add_column("Service Name")
            table.add_column("Region")
            table.add_column("URL")
            table.add_column("Proxied")
            table.add_column("Port")
            
            for service in services:
                service_name = service['metadata']['name']
                service_url = service['status']['url']
                if service_name in proxies:
                    proxied = "Yes"
                    proxy_port = proxies[service_name]['port']
                else:
                    proxied = "No"
                    proxy_port = "-"
                table.add_row(service_name, region, service_url, proxied, str(proxy_port))

            console.print(table)
        except Exception as e:
            status.stop()
            console.print(f"[bold red]ERROR: An unexpected error occurred: {e}[/bold red]")


def get_cloud_run_service_url(project, region, service_name):
    """
    Retrieves the URL of a specific Cloud Run service in a given project and region.

    Args:
        project (str): The GCP project ID.
        region (str): The region of the Cloud Run service.
        service_name (str): The name of the Cloud Run service.

    Returns:
        str: The URL of the Cloud Run service, or an error message if not found.
    """
        # Try to load existing data from the file, or initialize an empty dict
    file_path = "config/cloud_run_urls.json"
    if os.path.exists(file_path):
        with open(file_path, 'r') as file:
            services_dict = json.load(file)
    else:
        services_dict = {}

    try:
        result = subprocess.run(
            ["gcloud", "run", "services", "describe", service_name, "--project", project, "--region", region, "--format=json"],
            stdout=subprocess.PIPE, stderr=subprocess.PIPE, timeout=30
        )
        if result.returncode != 0:
            console.print(f"[bold red]ERROR: Unable to get Cloud Run service URL: {result.stderr.decode()}[/bold red]")
            return None

        service = json.loads(result.stdout.decode())
        service_url = service['status']['url']

        # Update the services dictionary
        services_dict[service_name] = service_url

        # Write the updated dictionary back to the file
        with open(file_path, 'w') as file:
            json.dump(services_dict, file, indent=4)
            
        return service_url
    except Exception as e:
        console.print(f"[bold red]ERROR: An unexpected error occurred: {e}[/bold red]")
        return None

    

def setup_vac_subparser(subparsers):
    """
    Sets up an argparse subparser for the 'vac' command.

    Args:
        subparsers: The subparsers object from argparse.ArgumentParser().
    """
    vac_parser = subparsers.add_parser('vac', help='Interact with deployed VAC services.')
    vac_parser.add_argument('--url_override', help='Override the VAC service URL.')
    vac_parser.add_argument('--no-proxy', action='store_true', help='Do not use the proxy and connect directly to the VAC service.')
    vac_subparsers = vac_parser.add_subparsers(dest='action', help='VAC subcommands')

    # Subcommand for listing VAC services
    list_parser = vac_subparsers.add_parser('list', help='List all VAC services.')

    # Subcommand for getting the URL of a specific VAC service
    get_url_parser = vac_subparsers.add_parser('get-url', help='Get the URL of a specific VAC service.')
    get_url_parser.add_argument('vac_name', help='Name of the VAC service.')

    # Subcommand for interacting with a VAC service
    chat_parser = vac_subparsers.add_parser('chat', help='Interact with a VAC service.')
    chat_parser.add_argument('vac_name', help='Name of the VAC service.')
    chat_parser.add_argument('user_input', help='User input for the VAC service when in headless mode.', nargs='?', default=None)
    chat_parser.add_argument('--headless', action='store_true', help='Run in headless mode.')
    chat_parser.add_argument('--chat_history', help='Chat history for headless mode (as JSON string).', default=None)

    # Subcommand for invoking a VAC service directly
    invoke_parser = vac_subparsers.add_parser('invoke', help='Invoke a VAC service directly with custom data.')
    invoke_parser.add_argument('vac_name', help='Name of the VAC service.')
    invoke_parser.add_argument('data', help='Data to send to the VAC service (as JSON string).')
    invoke_parser.add_argument('--is-file', action='store_true', help='Indicate if the data argument is a file path')

    # If no subcommand is provided, print the help message
    vac_parser.set_defaults(func=lambda args: vac_parser.print_help() if args.action is None else vac_command(args))



#   Copyright [2024] [Holosun ApS]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
MCP Server wrapper for VAC functionality.
This module exposes VAC streaming capabilities as MCP tools.
"""

from typing import Any, Sequence, Dict, List, Optional, Callable
import json
import asyncio
from functools import partial

from mcp.server import Server
from mcp.types import Tool, TextContent, ImageContent, EmbeddedResource

from ..custom_logging import log
from ..streaming import start_streaming_chat_async


class VACMCPServer:
    """MCP Server that exposes VAC functionality as tools."""
    
    def __init__(self, stream_interpreter: Callable, vac_interpreter: Callable = None):
        """
        Initialize the VAC MCP Server.
        
        Args:
            stream_interpreter: The streaming interpreter function
            vac_interpreter: The static VAC interpreter function (optional)
        """
        # MCP server is always available with current SDK
        
        self.stream_interpreter = stream_interpreter
        self.vac_interpreter = vac_interpreter
        self.server = Server("sunholo-vac-server")
        
        # Set up handlers
        self._setup_handlers()
    
    def _setup_handlers(self):
        """Set up MCP protocol handlers."""
        
        @self.server.list_tools()
        async def list_tools() -> List[Tool]:
            """List available VAC tools."""
            tools = [
                Tool(
                    name="vac_stream",
                    description="Stream responses from a Sunholo VAC (Virtual Agent Computer)",
                    inputSchema={
                        "type": "object",
                        "properties": {
                            "vector_name": {
                                "type": "string",
                                "description": "Name of the VAC to interact with"
                            },
                            "user_input": {
                                "type": "string", 
                                "description": "The user's question or input"
                            },
                            "chat_history": {
                                "type": "array",
                                "description": "Previous conversation history",
                                "items": {
                                    "type": "object",
                                    "properties": {
                                        "human": {"type": "string"},
                                        "ai": {"type": "string"}
                                    }
                                },
                                "default": []
                            },
                            "stream_wait_time": {
                                "type": "number",
                                "description": "Time to wait between stream chunks",
                                "default": 7
                            },
                            "stream_timeout": {
                                "type": "number",
                                "description": "Maximum time to wait for response",
                                "default": 120
                            }
                        },
                        "required": ["vector_name", "user_input"]
                    }
                )
            ]
            
            # Add static VAC tool if interpreter is provided
            if self.vac_interpreter:
                tools.append(
                    Tool(
                        name="vac_query",
                        description="Query a Sunholo VAC (non-streaming)",
                        inputSchema={
                            "type": "object",
                            "properties": {
                                "vector_name": {
                                    "type": "string",
                                    "description": "Name of the VAC to interact with"
                                },
                                "user_input": {
                                    "type": "string",
                                    "description": "The user's question or input"
                                },
                                "chat_history": {
                                    "type": "array",
                                    "description": "Previous conversation history",
                                    "items": {
                                        "type": "object",
                                        "properties": {
                                            "human": {"type": "string"},
                                            "ai": {"type": "string"}
                                        }
                                    },
                                    "default": []
                                }
                            },
                            "required": ["vector_name", "user_input"]
                        }
                    )
                )
            
            return tools
        
        @self.server.call_tool()
        async def call_tool(
            name: str,
            arguments: Any
        ) -> Sequence[TextContent | ImageContent | EmbeddedResource]:
            """Handle tool calls for VAC interactions."""
            
            if name == "vac_stream":
                return await self._handle_vac_stream(arguments)
            elif name == "vac_query" and self.vac_interpreter:
                return await self._handle_vac_query(arguments)
            else:
                raise ValueError(f"Unknown tool: {name}")
    
    async def _handle_vac_stream(self, arguments: Dict[str, Any]) -> Sequence[TextContent]:
        """Handle streaming VAC requests."""
        vector_name = arguments.get("vector_name")
        user_input = arguments.get("user_input")
        chat_history = arguments.get("chat_history", [])
        stream_wait_time = arguments.get("stream_wait_time", 7)
        stream_timeout = arguments.get("stream_timeout", 120)
        
        if not vector_name or not user_input:
            raise ValueError("Missing required arguments: vector_name and user_input")
        
        log.info(f"MCP streaming request for VAC '{vector_name}': {user_input}")
        
        try:
            # Collect streaming responses
            full_response = ""
            
            # Check if stream_interpreter is async
            if asyncio.iscoroutinefunction(self.stream_interpreter):
                async for chunk in start_streaming_chat_async(
                    question=user_input,
                    vector_name=vector_name,
                    qna_func_async=self.stream_interpreter,
                    chat_history=chat_history,
                    wait_time=stream_wait_time,
                    timeout=stream_timeout
                ):
                    if isinstance(chunk, dict) and 'answer' in chunk:
                        full_response = chunk['answer']
                    elif isinstance(chunk, str):
                        full_response += chunk
            else:
                # Fall back to sync version for non-async interpreters
                result = self.stream_interpreter(
                    question=user_input,
                    vector_name=vector_name,
                    chat_history=chat_history
                )
                if isinstance(result, dict):
                    full_response = result.get("answer", str(result))
                else:
                    full_response = str(result)
            
            return [
                TextContent(
                    type="text",
                    text=full_response or "No response generated"
                )
            ]
            
        except Exception as e:
            log.error(f"Error in MCP VAC stream: {str(e)}")
            return [
                TextContent(
                    type="text",
                    text=f"Error: {str(e)}"
                )
            ]
    
    async def _handle_vac_query(self, arguments: Dict[str, Any]) -> Sequence[TextContent]:
        """Handle non-streaming VAC requests."""
        vector_name = arguments.get("vector_name")
        user_input = arguments.get("user_input")
        chat_history = arguments.get("chat_history", [])
        
        if not vector_name or not user_input:
            raise ValueError("Missing required arguments: vector_name and user_input")
        
        log.info(f"MCP query request for VAC '{vector_name}': {user_input}")
        
        try:
            # Run in executor if not async
            if asyncio.iscoroutinefunction(self.vac_interpreter):
                result = await self.vac_interpreter(
                    question=user_input,
                    vector_name=vector_name,
                    chat_history=chat_history
                )
            else:
                loop = asyncio.get_event_loop()
                result = await loop.run_in_executor(
                    None,
                    partial(
                        self.vac_interpreter,
                        question=user_input,
                        vector_name=vector_name,
                        chat_history=chat_history
                    )
                )
            
            # Extract answer from result
            if isinstance(result, dict):
                answer = result.get("answer", str(result))
            else:
                answer = str(result)
            
            return [
                TextContent(
                    type="text",
                    text=answer
                )
            ]
            
        except Exception as e:
            log.error(f"Error in MCP VAC query: {str(e)}")
            return [
                TextContent(
                    type="text",
                    text=f"Error: {str(e)}"
                )
            ]
    
    def get_server(self) -> Server:
        """Get the underlying MCP server instance."""
        return self.server


from typing import Optional, List, Union, Dict, Any, TypedDict, TYPE_CHECKING, Generator

import enum
import json
from pydantic import BaseModel
import time
try:
    from google import genai
    from google.genai import types
except ImportError:
    genai = None

try:
    import sounddevice as sd
except ImportError:
    sd = None
except OSError:
    sd = None

try:
    import numpy as np
except ImportError:
    np = None

try:
    import cv2 as cv2
except ImportError:
    cv2 = None

if TYPE_CHECKING:
    from google import genai
    from google.genai import types
    from google.genai.types import Tool, GenerateContentConfig, EmbedContentConfig
else:
    genai = None

class GoogleAIConfig(BaseModel):
    """Configuration class for GoogleAI client initialization.
    See https://ai.google.dev/gemini-api/docs/models/gemini-v2
    """
    api_key: Optional[str] = None
    project_id: Optional[str] = None
    location: str = "us-central1"
    use_vertex: bool = False

class GoogleAI:
    """A wrapper class for Google's v2 Generative AI APIs.
    See https://ai.google.dev/gemini-api/docs/models/gemini-v2
    """
    
    def __init__(self, config: GoogleAIConfig):
        """Initialize the GoogleAI client.
        
        Args:
            config (GoogleAIConfig): Configuration for client initialization
        """
        if genai is None:
            raise ImportError("GoogleAI requires google-genai to be installed, try sunholo[gcp]")
        if config.use_vertex:
            if not config.project_id:
                raise ValueError("project_id is required for Vertex AI")
            self.client = genai.Client(
                vertexai=True,
                project=config.project_id,
                location=config.location
            )
        else:
            if not config.api_key:
                raise ValueError("api_key is required for Google AI API")
            self.client = genai.Client(api_key=config.api_key)
        
        self.default_model = "gemini-2.0-flash-exp"
    
    def google_search_tool(self) -> "types.Tool":
        from google.genai.types import Tool, GoogleSearch
        return Tool(
            google_search = GoogleSearch()
        )
    def generate_text(
        self,
        prompt: str,
        model: Optional[str] = None,
        temperature: float = 0.7,
        max_output_tokens: int = 1024,
        top_p: float = 0.95,
        top_k: int = 20,
        stop_sequences: Optional[List[str]] = None,
        system_prompt: Optional[str] = None,
        tools: Optional[List["types.Tool"]] = None
    ) -> str:
        """Generate text using the specified model.
        
        Args:
            prompt (str): The input prompt
            model (Optional[str]): Model name to use
            temperature (float): Controls randomness (0.0-1.0)
            max_output_tokens (int): Maximum number of tokens to generate
            top_p (float): Nucleus sampling parameter
            top_k (int): Top-k sampling parameter
            stop_sequences (Optional[List[str]]): Sequences that stop generation
            system_prompt (Optional[str]): System-level instruction
            tools: list of python functions or Tool objects
            
        Returns:
            str: Generated text response
        """
        model = model or self.default_model
        
        config = types.GenerateContentConfig(
            temperature=temperature,
            max_output_tokens=max_output_tokens,
            top_p=top_p,
            top_k=top_k,
            stop_sequences=stop_sequences or [],
            tools=tools or []
        )
        
        if system_prompt:
            config.system_instruction = system_prompt
            
        response = self.client.models.generate_content(
            model=model,
            contents=prompt,
            config=config
        )
        
        return response.text

    async def generate_text_async(
        self,
        prompt: str,
        model: Optional[str] = None,
        **kwargs
    ) -> str:
        """Async version of generate_text."""
        model = model or self.default_model
        response = await self.client.aio.models.generate_content(
            model=model,
            contents=prompt,
            config=types.GenerateContentConfig(**kwargs)
        )
        return response.text
        
    async def _record_audio(
            self,
            duration: float = 5.0,
            sample_rate: int = 16000
        ) -> bytes:
        """Internal method to record audio."""
        if sd is None or np is None:
            raise ImportError("sounddevice and numpy are required for audio. Install with pip install sunholo[tts]")
        
        print(f"Recording for {duration} seconds...")
        audio_data = sd.rec(
            int(duration * sample_rate),
            samplerate=sample_rate,
            channels=1,
            dtype=np.int16
        )
        sd.wait()
        print("Recording complete")
        return audio_data.tobytes()

    async def _record_video(
            self,
            duration: float = 5.0
        ) -> List[bytes]:
        """Internal method to record video frames."""
        import cv2
        
        frames = []
        screen = cv2.VideoCapture(0)
        start_time = time.time()
        
        try:
            while time.time() - start_time < duration:
                ret, frame = screen.read()
                if ret:
                    _, buffer = cv2.imencode('.jpg', frame)
                    frames.append(buffer.tobytes())
                    time.sleep(0.1)  # Limit frame rate
        finally:
            screen.release()
        
        return frames

    async def _process_responses(self, session) -> List[str]:
        """Internal method to process session responses."""
        responses = []
        i = 1
        async for response in session.receive():
            model_turn = response.server_content.model_turn
            if model_turn is None:
                continue
            for part in model_turn.parts:
                text = part.text
                print(f"[{i}] {text}")
                i += 1
                responses.append(text)
        return responses

    async def live_async(
            self,
            prompt: Optional[Union[str, List[Union[str, bytes]]]] = None,
            input_type: str = "text",  # "text", "audio", or "video"
            duration: Optional[float] = None,  # For audio/video recording duration
            model: Optional[str] = None,
            **kwargs
        ) -> str:
        """Live Multimodal API with support for text, audio, and video inputs.
        
        Args:
            input_type: Type of input ("text", "audio", or "video")
            prompt: Text prompt or list of text/binary chunks
            duration: Recording duration for audio/video in seconds
            model: Optional model name
            **kwargs: Additional configuration parameters
        
        Returns:
            str: Generated response text
        """
        client = genai.Client(
            http_options={
                'api_version': 'v1alpha',
                'url': 'generativelanguage.googleapis.com',
            }
        )
        
        config = {
            "generation_config": {"response_modalities": ["TEXT"]}
        }

        async with client.aio.live.connect(model=self.default_model, config=config) as session:
            # Handle different input types
            if input_type == "text":
                message = {
                    "client_content": {
                        "turns": [
                            {
                                "parts": [{"text": prompt}],
                                "role": "user"
                            }
                        ],
                        "turn_complete": True
                    }
                }
                await session.send(json.dumps(message), end_of_turn=True)
            
            elif input_type == "audio":
                audio_data = await self._record_audio(duration=duration or 5.0)
                message = {"media_chunks": [audio_data]}
                await session.send(message)
                await session.send(json.dumps({"turn_complete": True}), end_of_turn=True)
            
            elif input_type == "video":
                frames = await self._record_video(duration=duration or 5.0)
                for frame in frames:
                    message = {"media_chunks": [frame]}
                    await session.send(message)
                await session.send(json.dumps({"turn_complete": True}), end_of_turn=True)
            
            else:
                raise ValueError(f"Unsupported input_type: {input_type}")

            # Process responses
            responses = await self._process_responses(session)
            return "OK"
        
    def gs_uri(self, uri, mime_type=None):
      
        if mime_type is None:
            from ..utils.mime import guess_mime_type
            mime_type = guess_mime_type(uri)

        return types.Part.from_uri(
            file_uri=uri,
            mime_type=mime_type,
        )

    def local_file(self, filename, mime_type=None):
        if mime_type is None:
            from ..utils.mime import guess_mime_type
            mime_type = guess_mime_type(filename)

        with open(filename, 'rb') as f:
            image_bytes = f.read()
        
        if image_bytes and mime_type:
            return types.Part.from_bytes(
                data=image_bytes,
                mime_type=mime_type,
            )
        else:
            raise ValueError(f"Could not read bytes or mime_type for {filename=} - {mime_type=}")

    def stream_text(
        self,
        prompt: str,
        model: Optional[str] = None,
        **kwargs
    ) -> "Generator[str, None, None]":
        """Stream text generation responses.
        
        Args:
            prompt (str): The input prompt
            model (Optional[str]): Model name to use
            **kwargs: Additional configuration parameters
            
        Yields:
            str: Chunks of generated text
        """
        model = model or self.default_model
        for chunk in self.client.models.generate_content_stream(
            model=model,
            contents=prompt,
            config=types.GenerateContentConfig(**kwargs)
        ):
            yield chunk.text
            
    def get_embedding(
        self,
        text: Union[str, List[str]],
        model: str = "text-embedding-004",
        output_dim: Optional[int] = None
    ) -> Union[List[float], List[List[float]]]:
        """Get text embeddings.
        
        Args:
            text (Union[str, List[str]]): Text to embed
            model (str): Embedding model to use
            output_dim (Optional[int]): Desired embedding dimension
            
        Returns:
            Union[List[float], List[List[float]]]: Embeddings
        """
        config = {}
        if output_dim:
            config["output_dimensionality"] = output_dim
            
        response = self.client.models.embed_content(
            model=model,
            contents=text,
            config=types.EmbedContentConfig(**config) if config else None
        )
        
        if isinstance(text, str):
            return response.embeddings[0].values
        return [emb.values for emb in response.embeddings]

    def structured_output(
        self,
        prompt: str,
        schema: Union[BaseModel, Dict, type, TypedDict],
        model: Optional[str] = None,
        is_list: bool = False
    ) -> Dict:
        """Generate structured output according to a schema.
        
        Args:
            prompt (str): Input prompt
            schema (Union[BaseModel, Dict, type]): Schema definition (Pydantic model, TypedDict, or raw schema)
            model (Optional[str]): Model to use
            is_list (bool): Whether to wrap the schema in a list
            
        Returns:
            Dict: Structured response matching schema
        """
        model = model or self.default_model
        
        # Handle TypedDict differently than Pydantic models
        if isinstance(schema, type):
            if hasattr(schema, '__origin__') and schema.__origin__ is list:
                # Handle list[TypedDict] case
                inner_type = schema.__args__[0]
                config = types.GenerateContentConfig(
                    response_mime_type='application/json',
                    response_schema=schema
                )
            elif hasattr(schema, '__annotations__'):  # TypedDict check
                # Create API-compatible schema
                schema_dict = {
                    'type': 'OBJECT',
                    'properties': {},
                    'required': list(schema.__annotations__.keys())  # TypedDict fields are required by default
                }
                
                for field_name, field_type in schema.__annotations__.items():
                    if hasattr(field_type, '__base__') and field_type.__base__ == enum.Enum:
                        schema_dict['properties'][field_name] = {
                            'type': 'STRING',
                            'enum': [e.value for e in field_type]
                        }
                    elif field_type is str:
                        schema_dict['properties'][field_name] = {'type': 'STRING'}
                    elif field_type is int:
                        schema_dict['properties'][field_name] = {'type': 'INTEGER'}
                    elif field_type is float:
                        schema_dict['properties'][field_name] = {'type': 'NUMBER'}
                    elif field_type is bool:
                        schema_dict['properties'][field_name] = {'type': 'BOOLEAN'}
                    else:
                        schema_dict['properties'][field_name] = {'type': 'STRING'}
                
                if is_list:
                    schema_dict = {
                        'type': 'ARRAY',
                        'items': schema_dict
                    }
                
                config = types.GenerateContentConfig(
                    response_mime_type='application/json',
                    response_schema=schema_dict
                )
            elif hasattr(schema, '__origin__') and schema.__origin__ is list:
                # Handle List[TypeDict] case
                inner_type = schema.__args__[0]
                if hasattr(inner_type, '__annotations__'):  # Check if inner type is TypedDict
                    # Create API-compatible schema for the inner TypedDict
                    schema_dict = {
                        'type': 'OBJECT',
                        'properties': {},
                        'required': list(inner_type.__annotations__.keys())
                    }
                    
                    for field_name, field_type in inner_type.__annotations__.items():
                        if hasattr(field_type, '__base__') and field_type.__base__ == enum.Enum:
                            schema_dict['properties'][field_name] = {
                                'type': 'STRING',
                                'enum': [e.value for e in field_type]
                            }
                        elif field_type is str:
                            schema_dict['properties'][field_name] = {'type': 'STRING'}
                        elif field_type is int:
                            schema_dict['properties'][field_name] = {'type': 'INTEGER'}
                        elif field_type is float:
                            schema_dict['properties'][field_name] = {'type': 'NUMBER'}
                        elif field_type is bool:
                            schema_dict['properties'][field_name] = {'type': 'BOOLEAN'}
                        else:
                            schema_dict['properties'][field_name] = {'type': 'STRING'}
                    
                    # Wrap in array type
                    array_schema = {
                        'type': 'ARRAY',
                        'items': schema_dict
                    }
                    
                    config = types.GenerateContentConfig(
                        response_mime_type='application/json',
                        response_schema=array_schema
                    )
            elif issubclass(schema, BaseModel):
                # Convert Pydantic model to Google's schema format
                schema_dict = {
                    'type': 'OBJECT',
                    'properties': {},
                    'required': []
                }
                
                for field_name, field_info in schema.model_fields.items():
                    # Handle different field types including enums
                    if isinstance(field_info.annotation, type) and issubclass(field_info.annotation, enum.Enum):
                        field_type = {
                            'type': 'STRING',
                            'enum': [e.value for e in field_info.annotation]
                        }
                    elif field_info.annotation is str:
                        field_type = {'type': 'STRING'}
                    elif field_info.annotation is int:
                        field_type = {'type': 'INTEGER'}
                    elif field_info.annotation is float:
                        field_type = {'type': 'NUMBER'}
                    elif field_info.annotation is bool:
                        field_type = {'type': 'BOOLEAN'}
                    elif field_info.annotation is list or (
                        hasattr(field_info.annotation, '__origin__') and 
                        field_info.annotation.__origin__ is list
                    ):
                        # Handle typed lists
                        if hasattr(field_info.annotation, '__args__'):
                            inner_type = field_info.annotation.__args__[0]
                            if inner_type is str:
                                item_type = 'STRING'
                            elif inner_type is int:
                                item_type = 'INTEGER'
                            elif inner_type is float:
                                item_type = 'NUMBER'
                            elif inner_type is bool:
                                item_type = 'BOOLEAN'
                            else:
                                item_type = 'STRING'
                        else:
                            item_type = 'STRING'
                        field_type = {'type': 'ARRAY', 'items': {'type': item_type}}
                    else:
                        field_type = {'type': 'STRING'}
                    
                    schema_dict['properties'][field_name] = field_type
                    if field_info.is_required:
                        schema_dict['required'].append(field_name)
                
                if is_list:
                    schema_dict = {
                        'type': 'ARRAY',
                        'items': schema_dict
                    }
                
                config = types.GenerateContentConfig(
                    response_mime_type='application/json',
                    response_schema=schema_dict
                )
        else:
            # Handle raw schema dict
            config = types.GenerateContentConfig(
                response_mime_type='application/json',
                response_schema=schema
            )
            
        response = self.client.models.generate_content(
            model=model,
            contents=prompt,
            config=config
        )
        
        return response.text
    def count_tokens(self, text: str, model: Optional[str] = None) -> int:
        """Count the number of tokens in the text.
        
        Args:
            text (str): Input text
            model (Optional[str]): Model to use for tokenization
            
        Returns:
            int: Number of tokens
        """
        model = model or self.default_model
        response = self.client.models.count_tokens(
            model=model,
            contents=text
        )
        return response.total_tokens




"""
Proper MCP integration for VACRoutes using the official MCP Python SDK.
This shows how to integrate MCP servers with your Flask/VACRoutes application.
"""

from typing import Dict, Any, List, Optional
import asyncio

# MCP SDK imports - try different import paths
try:
    from mcp.client.stdio import StdioClientTransport
    from mcp.client.session import ClientSession
except ImportError:
    try:
        # Alternative import paths
        from mcp.client import StdioClientTransport, ClientSession
    except ImportError:
        try:
            # Another alternative
            from mcp import StdioClientTransport, ClientSession
        except ImportError:
            StdioClientTransport = None
            ClientSession = None

try:
    from mcp.types import Tool, Resource, TextContent, CallToolResult
except ImportError:
    try:
        from mcp import Tool, Resource, TextContent, CallToolResult
    except ImportError:
        Tool = None
        Resource = None
        TextContent = None
        CallToolResult = None


class MCPClientManager:
    """Manages MCP client connections to various MCP servers."""
    
    def __init__(self):
        self.sessions: Dict[str, ClientSession] = {}
        self.server_configs: Dict[str, Dict[str, Any]] = {}
        
    async def connect_to_server(self, server_name: str, command: str, args: List[str] = None) -> ClientSession:
        """Connect to an MCP server via stdio."""
        if server_name in self.sessions:
            return self.sessions[server_name]
        
        if not StdioClientTransport or not ClientSession:
            raise ImportError("MCP client dependencies not available")
        
        # Create transport and session
        transport = StdioClientTransport(
            command=command,
            args=args or []
        )
        session = ClientSession(transport)
        await session.initialize()
        
        self.sessions[server_name] = session
        self.server_configs[server_name] = {
            "command": command,
            "args": args
        }
        return session
    
    async def list_tools(self, server_name: Optional[str] = None) -> List[Tool]:
        """List available tools from one or all connected servers."""
        if server_name:
            session = self.sessions.get(server_name)
            if session:
                result = await session.list_tools()
                return result.tools
            return []
        
        # List from all servers
        all_tools = []
        for name, session in self.sessions.items():
            result = await session.list_tools()
            # Add server name to tool metadata
            for tool in result.tools:
                tool.metadata = tool.metadata or {}
                tool.metadata["server"] = name
            all_tools.extend(result.tools)
        return all_tools
    
    async def call_tool(self, server_name: str, tool_name: str, arguments: Dict[str, Any]) -> CallToolResult:
        """Call a tool on a specific MCP server."""
        session = self.sessions.get(server_name)
        if not session:
            raise ValueError(f"Not connected to server: {server_name}")
        
        # Call the tool
        try:
            from mcp.types import CallToolRequest
            request = CallToolRequest(name=tool_name, arguments=arguments)
            result = await session.call_tool(request)
        except ImportError:
            # Try direct call if Request types not available
            result = await session.call_tool(tool_name, arguments)
        return result
    
    async def list_resources(self, server_name: Optional[str] = None) -> List[Resource]:
        """List available resources from servers."""
        if server_name:
            session = self.sessions.get(server_name)
            if session:
                result = await session.list_resources()
                return result.resources
            return []
        
        # List from all servers
        all_resources = []
        for name, session in self.sessions.items():
            result = await session.list_resources()
            for resource in result.resources:
                resource.metadata = resource.metadata or {}
                resource.metadata["server"] = name
            all_resources.extend(result.resources)
        return all_resources
    
    async def read_resource(self, server_name: str, uri: str) -> List[TextContent]:
        """Read a resource from an MCP server."""
        session = self.sessions.get(server_name)
        if not session:
            raise ValueError(f"Not connected to server: {server_name}")
        
        try:
            from mcp.types import ReadResourceRequest
            request = ReadResourceRequest(uri=uri)
            result = await session.read_resource(request)
        except ImportError:
            # Try direct call if Request types not available
            result = await session.read_resource(uri)
        
        return result.contents if hasattr(result, 'contents') else result


from ..custom_logging import log
from ..gcs import get_bytes_from_gcs

from functools import partial
import mimetypes
import uuid
import asyncio
import tempfile
import re
import os
import traceback
try:
    import google.generativeai as genai
    from google import genai as genaiv2
    from google.genai import types

except ImportError:
    genai = None
    genaiv2 = None

DOCUMENT_MIMES = [
    'application/pdf',
    'application/x-javascript',
    'text/javascript',
    'application/x-python',
    'text/x-python',
    'text/plain',
    'text/html',
    'text/css',
    'text/md',
    'text/csv',
    'text/xml',
    'text/rtf'
    ]

IMAGE_MIMES = [
    'image/png',
    'image/jpeg',
    'image/webp',
    'image/heic',
    'image/heif',
]

VIDEO_MIMES = [
    'video/mp4',
    'video/mpeg',
    'video/mov',
    'video/avi',
    'video/x-flv',
    'video/mpg',
    'video/webm',
    'video/wmv',
    'video/3gpp'
]

AUDIO_MIMES = [
    'audio/wav',
    'audio/mp3',
    'audio/mpeg', #added
    'audio/aiff',
    'audio/aac',
    'audio/ogg',
    'audio/flac',    
]

ALLOWED_MIME_TYPES = set(AUDIO_MIMES + VIDEO_MIMES + IMAGE_MIMES + DOCUMENT_MIMES)

# 'documents': 
# [
#   {'storagePath': 'users/UQcKi4u7s...dsd.png', 
#    'url': 'https://firebasestorage.googleapis.com/v0/b/multi...', 
#    'contentType': 'image/png', 
#    'type': 'image', 
#    'name': 'multivac-data-architecture.png'}, 
#   {'storagePath': 'users/UQc...3dc59e1.jpg', 
#   'type': 'image', 
#   'name': 'holosun-circle.jpg', 
#   'url': 'https://firebasestorage.googleapis.com/v0/b/multiv...', 
#   'contentType': 'image/jpeg'}
# ]

def sanitize_file(filename):
    """
    Sanitize filename to conform to Gemini API requirements:
    - Only lowercase alphanumeric characters and dashes
    - Cannot begin or end with a dash
    """
    # Split the filename into name and extension
    name, extension = os.path.splitext(filename)
    
    # Convert to lowercase
    name = name.lower()
    
    # Replace any non-alphanumeric characters (including underscores) with dashes
    sanitized_name = re.sub(r'[^a-z0-9]', '-', name)
    
    # Remove consecutive dashes
    sanitized_name = re.sub(r'-+', '-', sanitized_name)
    
    # Remove leading or trailing dashes
    sanitized_name = sanitized_name.strip('-')
    
    # If the name is empty after sanitization, use a default
    if not sanitized_name:
        sanitized_name = 'file'
    
    # Limit the length
    return sanitized_name[:40]

async def construct_file_content(gs_list, 
                                 bucket:str, 
                                 genai_lib=False, 
                                 timeout=60,
                                 genai_lib_vertex=False):
    """
    Args:
    - gs_list: a list of dicts representing files in a bucket
    - bucket: The bucket the files are in
    - genai_lib: whether its using the genai SDK
    - timeout: timeout in seconds for the gather operation
    - genai_lib_vertex: bool whether to use the vertex version of the genai library, using default credentials
    """
    
    file_list = []
    for element in gs_list:
        the_mime_type = element.get('contentType')
        if the_mime_type is None:
            continue
        if element.get('storagePath') is None:
            continue
        if the_mime_type in ALLOWED_MIME_TYPES:
            file_list.append(element)
        else:
            log.warning(f'{the_mime_type} is not in allowed MIME types for {element.get("name")}')
    
    if not file_list:
        return [{"role": "user", "parts": [{"text": "No eligible contentTypes were found"}]}]

    # export GOOGLE_GENAI_USE_VERTEXAI=true
    # export GOOGLE_CLOUD_PROJECT='your-project-id'
    # export GOOGLE_CLOUD_LOCATION='us-central1'
    content = []
    if genai_lib_vertex or os.getenv("GOOGLE_GENAI_USE_VERTEXAI"):
        client = genaiv2.Client(vertexai=True)
        # we can just use the gs:// uri directly
        for file_info in file_list:
            img_url = f"gs://{bucket}/{file_info['storagePath']}"
            mime_type = file_info['contentType']
            display_name = file_info['name']
            display_url = img_url
            file_content = types.Part.from_uri(
                file_uri=img_url,
                mime_type=mime_type,
            )
            file_content_list = [file_content, 
                f"You have been given the ability to read and work with filename '{display_name}' with {mime_type=} {display_url=}"] 
            content.extend(file_content_list)

        return content       

    
    # Loop through the valid files and process them
    tasks = []
    for file_info in file_list:
        img_url = f"gs://{bucket}/{file_info['storagePath']}"
        display_url = file_info.get('url')
        mime_type = file_info['contentType']
        # Generate a unique name for each file
        unique_name= sanitize_file(file_info['name'])

        display_name = file_info['name']
        log.info(f"Processing {unique_name=} {display_name=}")
        
        try:
            if not genai_lib:
                myfile = genai.get_file(unique_name)
            else:
                client = genaiv2.Client()
                myfile = client.files.get(name=unique_name)
            content.append(myfile)
            content.append(f"You have been given the ability to work with file {display_name=} with {mime_type=} {display_url=}")
            log.info(f"Found existing genai.get_file {unique_name=}")
        except Exception as e:
            log.info(f"Not found checking genai.get_file: '{unique_name}' {str(e)}")
            tasks.append(
                download_gcs_upload_genai(img_url, 
                                         mime_type=mime_type, 
                                         name=unique_name, 
                                         display_url=display_url, 
                                         display_name=display_name,
                                         genai_lib=genai_lib)
                                        )

    # Process files with timeout
    if tasks:
        try:
            # Add timeout to prevent hanging
            task_results = await asyncio.wait_for(asyncio.gather(*tasks, return_exceptions=True), timeout=timeout)
            
            # Process results, handle any exceptions
            for result in task_results:
                if isinstance(result, Exception):
                    log.error(f"Task failed with error: {str(result)}")
                    # Add error message to content
                    content.append({"role": "user", "parts": [{"text": f"Error processing file: {str(result)}"}]})
                else:
                    # Normalize the result structure based on genai_lib
                    if genai_lib and isinstance(result, list):
                        content.append(result[0])  # The file object
                        content.append({"role": "user", "parts": [{"text": result[1]}]})  # The message
                    else:
                        content.append(result)
        except asyncio.TimeoutError:
            log.error(f"Timeout occurred after {timeout} seconds while processing files")
            content.append({"role": "user", "parts": [{"text": f"Some files could not be processed within the time limit ({timeout} seconds)"}]})
    
    log.info(f"construct_file_content completed with {len(content)} items")
    return content

# Helper function to handle each file download with error handling
async def download_file_with_error_handling(img_url, mime_type, name):
    try:
        return await download_gcs_upload_genai(img_url, mime_type, name)
    except Exception as err:
        msg= f"Error processing file from {img_url}: {str(err)}"
        log.error(msg)
        return {"role": "user", "parts": [{"text": msg}]}

async def download_gcs_upload_genai(img_url, 
                                    mime_type, 
                                    name=None, 
                                    display_url=None, 
                                    display_name=None, 
                                    retries=3, delay=2, 
                                    genai_lib=False):
    """
    Downloads and uploads a file with retries in case of failure.
    Thread-safe implementation using unique file paths.
    
    Args:
      - img_url: str The URL of the file to download.
      - mime_type: str The MIME type of the file.
      - name: str Optional name, else a random one will be created
      - retries: int Number of retry attempts before failing.
      - delay: int Initial delay between retries, exponentially increasing.
      - genai_lin: bool whether to use newer genai library to upload files

      
    Returns:
      - downloaded_content: The result of the file upload if successful.
    """
    import aiofiles
    for attempt in range(retries):
        try:
            log.info(f"Upload attempt [{attempt}] for {img_url=}")
            # Download the file bytes asynchronously
            file_bytes = await asyncio.to_thread(get_bytes_from_gcs, img_url)
            if not file_bytes:
                msg = f"Failed to download file from {img_url}: got None"
                log.warning(msg)
                return {"role": "user", "parts": [{"text": msg}]}
            
            # Log the size of the file bytes
            file_size = len(file_bytes)
            log.info(f"Downloaded file size for {img_url}: {file_size} bytes")

            if file_size > 19434343:
                log.warning(f"File size for {img_url}: {file_size} is too big.")
                msg = f"The file for {img_url} is too large ({file_size} bytes) to be used directly. Use RAG instead or {display_url=}"
                return {"role": "user", "parts": [{"text": msg}]}
            
            extension = mimetypes.guess_extension(mime_type)
            
            # Create a unique directory for this upload task
            unique_id = str(uuid.uuid4())
            temp_dir = os.path.join(tempfile.gettempdir(), f"upload_{unique_id}")
            os.makedirs(temp_dir, exist_ok=True)

            # Create a file with unique path - NO sanitization in the filesystem path
            file_path = os.path.join(temp_dir, f"file_{unique_id}{extension}")
            
            api_name = sanitize_file(f"file-{unique_id[:8]}")


            log.info(f"Writing file {file_path}")
            async with aiofiles.open(file_path, 'wb') as f:
                await f.write(file_bytes)

            try:
                if not genai_lib:
                    downloaded_content = await asyncio.to_thread(
                            partial(genai.upload_file, name=api_name, mime_type=mime_type, display_name=display_name), 
                            file_path  # Use the unsanitized path for filesystem access
                        )
                    
                    # Clean up after successful upload
                    try:
                        os.remove(file_path)
                        os.rmdir(temp_dir)
                    except OSError as e:
                        log.warning(f"Cleanup error (non-critical): {str(e)}")
                        
                    return {"role": "user", "parts": [{"file_data": downloaded_content}, 
                                                    {"text": f"You have been given the ability to read and work with filename '{display_name}' with {mime_type=} {display_url=}"}
                                                   ]}

                else:

                    client = genaiv2.Client()
                    
                    # Use semaphore to limit concurrent uploads
                    async with upload_semaphore:
                        downloaded_content = await asyncio.to_thread(
                            client.files.upload, 
                            path=file_path,  # Use the unsanitized path for filesystem access
                            config=dict(mime_type=mime_type, display_name=display_name, name=api_name)
                        )
                    
                    # Clean up after successful upload
                    try:
                        os.remove(file_path)
                        os.rmdir(temp_dir)
                    except OSError as e:
                        log.warning(f"Cleanup error (non-critical): {str(e)}")
                        
                    return [downloaded_content, 
                            f"You have been given the ability to read and work with filename '{display_name}' with {mime_type=} {display_url=}"]

            except Exception as err:
                # Clean up on error
                try:
                    os.remove(file_path)
                    os.rmdir(temp_dir)
                except OSError:
                    pass
                
                msg = f"Could not upload {file_path} to {'genai.upload_file' if not genai_lib else 'genaiv2.client.files.upload'}: {str(err)} {traceback.format_exc()} {display_url=}"
                log.error(msg)
                return {"role": "user", "parts": [{"text": msg}]}
        
        except Exception as err:
            log.error(f"Error processing file {img_url} {mime_type=} on attempt {attempt + 1}/{retries}: {str(err)}")

            if attempt < retries - 1:
                log.info(f"Retrying in {delay} seconds...")
                await asyncio.sleep(delay)
                delay *= 2  # Exponential backoff
            else:
                raise err  # Raise the error after max retries

# Add this at the module level
# Create a semaphore to limit concurrent uploads
upload_semaphore = asyncio.Semaphore(5)  # Adjust the value based on your needs




#   Copyright [2024] [Holosun ApS]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from ...custom_logging import log
import time

def create_app(name):
    from flask import Flask, request
    
    app = Flask(name)

    app.config['TRAP_HTTP_EXCEPTIONS'] = True
    app.config['PROPAGATE_EXCEPTIONS'] = True
   
    @app.before_request
    def start_timer():
        request.start_time = time.time()

    @app.after_request  
    def log_timing(response):
        if hasattr(request, 'start_time'):
            duration = time.time() - request.start_time
            
            # Log all VAC requests with different detail levels
            if request.path.startswith('/vac/streaming/'):
                log.info(f"🚀 STREAMING: {duration:.3f}s - {request.path}")
            elif request.path.startswith('/vac/'):
                log.info(f"⚡ VAC: {duration:.3f}s - {request.path}")
            elif duration > 1.0:  # Log any slow requests
                log.warning(f"🐌 SLOW REQUEST: {duration:.3f}s - {request.path}")
            
            # Add response headers with timing info for debugging
            response.headers['X-Response-Time'] = f"{duration:.3f}s"
            
        return response
    
    return app



import os
import traceback

# app.py
from fastapi import FastAPI, Request
from fastapi.responses import JSONResponse

from my_log import log

app = FastAPI()

@app.get("/")
def home():
    """Simple endpoint to indicate that the app is running."""
    return {"message": "Hello, service!"}

@app.post("/system_service/<param>")
async def system_service(request: Request):
    """
    Pubsub message parsed and sent to Langfuse ID server
    """
    data = await request.json()

    try:
        #TODO: add stuff here
        meta = ""
        return {"status": "success", "message": meta}
    except Exception as err:
        log.error(f'EVAL_ERROR: Error when sending {data} to /pubsub_to_langfuse: {str(err)} traceback: {traceback.format_exc()}')
        return JSONResponse(status_code=200, content={"status": "error", "message": f'{str(err)} traceback: {traceback.format_exc()}'})

@app.post("/test_endpoint")
async def test_me(request: Request):
    """
    Endpoint to send trace_ids directly for evals then sent to Langfuse ID server
    """
    data = await request.json()

    try:
        #TODO: do something here
        meta = ""
        return {"status": "success", "message": meta}
    except Exception as err:
        log.error(f'EVAL_ERROR: Error when sending {data} to /direct_evals: {str(err)} traceback: {traceback.format_exc()}')
        return JSONResponse(status_code=500, content={"status": "error", "message": f'{str(err)} traceback: {traceback.format_exc()}'})

if __name__ == "__main__":
    import uvicorn
    uvicorn.run(app, host="0.0.0.0", port=int(os.environ.get("PORT", 8080)), debug=True)


#   Copyright [2024] [Holosun ApS]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Agent Card generation for A2A protocol.
Generates agent.json discovery documents for VACs.
"""

from typing import Dict, List, Any, Optional
import os
from ..utils import ConfigManager
from ..custom_logging import log
from ..utils.version import sunholo_version


class AgentCardGenerator:
    """Generates A2A Agent Cards for VAC discovery."""
    
    def __init__(self, base_url: str, config_manager: Optional[ConfigManager] = None):
        """
        Initialize the Agent Card generator.
        
        Args:
            base_url: The base URL where the A2A agent is hosted
            config_manager: Optional ConfigManager for global configs
        """
        self.base_url = base_url.rstrip('/')
        self.config_manager = config_manager or ConfigManager('global', validate=False)
        
    def generate_agent_card(self, vac_names: List[str] = None) -> Dict[str, Any]:
        """
        Generate the Agent Card JSON for A2A discovery.
        
        Args:
            vac_names: List of VAC names to include. If None, discovers all configured VACs.
            
        Returns:
            Dict containing the Agent Card JSON structure
        """
        if vac_names is None:
            vac_names = self._discover_vacs()
            
        skills = []
        agent_name = "Sunholo VAC Agent"
        agent_description = "Multi-VAC agent providing access to Sunholo Virtual Agent Computers"
        
        # Generate skills for each VAC
        for vac_name in vac_names:
            try:
                vac_config = ConfigManager(vac_name, validate=False)
                vac_skills = self._generate_vac_skills(vac_name, vac_config)
                skills.extend(vac_skills)
                
                # Use first VAC display name as primary agent name if available
                if len(skills) == len(vac_skills):  # First VAC
                    display_name = vac_config.vacConfig('display_name')
                    if display_name:
                        agent_name = f"{display_name} Agent"
                        
            except Exception as e:
                log.warning(f"Failed to process VAC {vac_name}: {e}")
                continue
        
        # If we have multiple VACs, use a generic name
        if len(vac_names) > 1:
            agent_name = "Sunholo Multi-VAC Agent"
            agent_description = f"Agent providing access to {len(vac_names)} Sunholo VACs: {', '.join(vac_names)}"
        
        agent_card = {
            "name": agent_name,
            "description": agent_description,
            "url": self.base_url,
            "version": sunholo_version(),
            "capabilities": [
                "task_management",
                "streaming",
                "conversation",
                "document_retrieval"
            ],
            "skills": skills,
            "metadata": {
                "framework": "sunholo",
                "protocol_version": "a2a-v1",
                "supported_formats": ["text", "json"],
                "max_context_length": self._get_max_context_length(vac_names),
                "languages": ["en"],  # TODO: Make configurable
                "created_at": self._get_current_timestamp()
            }
        }
        
        return agent_card
    
    def _discover_vacs(self) -> List[str]:
        """
        Discover all configured VACs from the configuration.
        
        Returns:
            List of VAC names found in the configuration
        """
        try:
            vac_configs = self.config_manager.configs_by_kind.get('vacConfig', {})
            if 'vac' in vac_configs:
                return list(vac_configs['vac'].keys())
            return []
        except Exception as e:
            log.error(f"Failed to discover VACs: {e}")
            return []
    
    def _generate_vac_skills(self, vac_name: str, vac_config: ConfigManager) -> List[Dict[str, Any]]:
        """
        Generate A2A skills for a specific VAC.
        
        Args:
            vac_name: Name of the VAC
            vac_config: ConfigManager instance for the VAC
            
        Returns:
            List of skill definitions for this VAC
        """
        skills = []
        
        # Get VAC metadata
        display_name = vac_config.vacConfig('display_name') or vac_name
        model = vac_config.vacConfig('model') or 'unknown'
        agent_type = vac_config.vacConfig('agent') or 'unknown'
        memory_config = vac_config.vacConfig('memory') or []
        
        # Skill 1: VAC Query (static response)
        query_skill = {
            "name": f"vac_query_{vac_name}",
            "description": f"Send a query to {display_name} and get a complete response using {model}",
            "input_schema": {
                "type": "object",
                "properties": {
                    "query": {
                        "type": "string",
                        "description": "The question or instruction to send to the VAC"
                    },
                    "chat_history": {
                        "type": "array",
                        "description": "Previous conversation history",
                        "items": {
                            "type": "object",
                            "properties": {
                                "role": {"type": "string", "enum": ["user", "assistant"]},
                                "content": {"type": "string"}
                            }
                        },
                        "default": []
                    },
                    "context": {
                        "type": "object",
                        "description": "Additional context or parameters",
                        "default": {}
                    }
                },
                "required": ["query"]
            },
            "output_schema": {
                "type": "object",
                "properties": {
                    "answer": {"type": "string"},
                    "source_documents": {"type": "array"},
                    "metadata": {"type": "object"}
                }
            },
            "metadata": {
                "vac_name": vac_name,
                "model": model,
                "agent_type": agent_type,
                "memory_enabled": len(memory_config) > 0,
                "response_type": "complete"
            }
        }
        skills.append(query_skill)
        
        # Skill 2: VAC Stream (streaming response)
        stream_skill = {
            "name": f"vac_stream_{vac_name}",
            "description": f"Start a streaming conversation with {display_name} using {model}",
            "input_schema": {
                "type": "object",
                "properties": {
                    "query": {
                        "type": "string",
                        "description": "The question or instruction to send to the VAC"
                    },
                    "chat_history": {
                        "type": "array",
                        "description": "Previous conversation history",
                        "items": {
                            "type": "object",
                            "properties": {
                                "role": {"type": "string", "enum": ["user", "assistant"]},
                                "content": {"type": "string"}
                            }
                        },
                        "default": []
                    },
                    "stream_settings": {
                        "type": "object",
                        "properties": {
                            "wait_time": {"type": "number", "default": 7},
                            "timeout": {"type": "number", "default": 120}
                        },
                        "default": {}
                    }
                },
                "required": ["query"]
            },
            "output_schema": {
                "type": "object",
                "properties": {
                    "stream_url": {"type": "string"},
                    "task_id": {"type": "string"},
                    "estimated_duration": {"type": "number"}
                }
            },
            "metadata": {
                "vac_name": vac_name,
                "model": model,
                "agent_type": agent_type,
                "memory_enabled": len(memory_config) > 0,
                "response_type": "streaming",
                "supports_sse": True
            }
        }
        skills.append(stream_skill)
        
        # Skill 3: Memory Search (if VAC has memory configured)
        if memory_config:
            memory_skill = {
                "name": f"vac_memory_search_{vac_name}",
                "description": f"Search through {display_name}'s knowledge base and memory",
                "input_schema": {
                    "type": "object",
                    "properties": {
                        "search_query": {
                            "type": "string",
                            "description": "The search query for the knowledge base"
                        },
                        "limit": {
                            "type": "integer",
                            "description": "Maximum number of results to return",
                            "default": 10,
                            "minimum": 1,
                            "maximum": 50
                        },
                        "similarity_threshold": {
                            "type": "number",
                            "description": "Minimum similarity score for results",
                            "default": 0.7,
                            "minimum": 0.0,
                            "maximum": 1.0
                        }
                    },
                    "required": ["search_query"]
                },
                "output_schema": {
                    "type": "object",
                    "properties": {
                        "results": {
                            "type": "array",
                            "items": {
                                "type": "object",
                                "properties": {
                                    "content": {"type": "string"},
                                    "score": {"type": "number"},
                                    "metadata": {"type": "object"}
                                }
                            }
                        },
                        "total_results": {"type": "integer"}
                    }
                }
            }
            skills.append(memory_skill)
        
        return skills
    
    def _get_max_context_length(self, vac_names: List[str]) -> int:
        """
        Determine the maximum context length based on configured models.
        
        Args:
            vac_names: List of VAC names to check
            
        Returns:
            Maximum context length in tokens
        """
        max_length = 8192  # Default
        
        for vac_name in vac_names:
            try:
                vac_config = ConfigManager(vac_name, validate=False)
                model = vac_config.vacConfig('model')
                
                # Map known models to context lengths
                if model:
                    if 'gemini-1.5' in model:
                        max_length = max(max_length, 2000000)  # Gemini 1.5 Pro
                    elif 'gpt-4' in model:
                        max_length = max(max_length, 128000)   # GPT-4 Turbo
                    elif 'claude-3' in model:
                        max_length = max(max_length, 200000)   # Claude 3
                    elif 'gemini' in model:
                        max_length = max(max_length, 32768)    # Other Gemini models
                        
            except Exception as e:
                log.debug(f"Could not determine context length for {vac_name}: {e}")
                continue
                
        return max_length
    
    def _get_current_timestamp(self) -> str:
        """Get current timestamp in ISO format."""
        from datetime import datetime
        return datetime.now().isoformat()
    
    def generate_discovery_endpoints(self) -> Dict[str, str]:
        """
        Generate the A2A discovery endpoint paths.
        
        Returns:
            Dictionary mapping endpoint names to their paths
        """
        return {
            "agent_card": "/.well-known/agent.json",
            "task_send": "/a2a/tasks/send",
            "task_send_subscribe": "/a2a/tasks/sendSubscribe",
            "task_get": "/a2a/tasks/get",
            "task_cancel": "/a2a/tasks/cancel",
            "task_notification": "/a2a/tasks/pushNotification/set"
        }


import os
import asyncio
from typing import Any, Sequence
from functools import lru_cache
import subprocess

try:
    from mcp.server import Server
    from mcp.server.stdio import stdio_server
    from mcp.types import (
        Resource,
        Tool,
        TextContent,
        ImageContent,
        EmbeddedResource,
    )

    from rich import print
    from ..cli.sun_rich import console
except ImportError:
    Server = None
    console = None

from pydantic import AnyUrl

# Configure logging
from ..custom_logging import setup_logging
logger = setup_logging("sunholo-mcp")

class SunholoMCPServer:
    def __init__(self):
        logger.info("Initializing Sunholo MCP Server")

        if Server is None:
            raise ImportError("SunholoMCPServer requires `sunholo[anthropic]` to be installed")
        
        self.server = Server("sunholo-mcp-server")
        self.server.onerror = self.handle_error

        self.setup_handlers()

    def handle_error(self, error: Exception):
        """Handle server errors"""
        logger.error(f"MCP Server error: {error}", exc_info=True)

    def setup_handlers(self):
        """Set up all the MCP protocol handlers"""
        self.setup_resource_handlers()
        self.setup_tool_handlers()

    def setup_resource_handlers(self):
        """Configure resource-related handlers"""
        
        @self.server.list_resources()
        async def list_resources() -> list[Resource]:
            """List available Sunholo resources"""
            return [
                Resource(
                    uri="sunholo://vacs/list",
                    name="Available Sunholo VACs",
                    mimeType="application/json",
                    description="List of available Virtual Agent Computers"
                )
            ]

        @self.server.read_resource()
        async def read_resource(uri: AnyUrl) -> str:
            """Read Sunholo resources based on URI"""
            logger.info(f"{uri} available")
            if str(uri) == "sunholo://vacs/list":
                try:
                    # Execute sunholo vac list command
                    result = subprocess.run(
                        ["sunholo", "vac", "list"],
                        capture_output=True,
                        text=True
                    )
                    return result.stdout
                except subprocess.CalledProcessError as e:
                    raise RuntimeError(f"Failed to list VACs: {str(e)}")
                except Exception as e:
                    raise RuntimeError(f"Error accessing Sunholo: {str(e)}")

            raise ValueError(f"Unknown resource: {uri}")

    def setup_tool_handlers(self):
        """Configure tool-related handlers"""

        @self.server.list_tools()
        async def list_tools() -> list[Tool]:
            """List available Sunholo tools"""
            return [
                Tool(
                    name="chat_with_vac",
                    description="Chat with a specific Sunholo VAC",
                    inputSchema={
                        "type": "object",
                        "properties": {
                            "vac_name": {
                                "type": "string",
                                "description": "Name of the VAC to chat with"
                            },
                            "message": {
                                "type": "string",
                                "description": "Message to send to the VAC"
                            },
                        },
                        "required": ["vac_name", "message"]
                    }
                ),
                Tool(
                    name="list_configs",
                    description="List Sunholo configurations",
                    inputSchema={
                        "type": "object",
                        "properties": {
                            "kind": {
                                "type": "string",
                                "description": "Filter configurations by kind e.g. vacConfig"
                            },
                            "vac": {
                                "type": "string", 
                                "description": "Filter configurations by VAC name"
                            },
                            "validate": {
                                "type": "boolean",
                                "description": "Validate the configuration files"
                            }
                        }
                    }
                ),
                Tool(
                    name="embed_content",
                    description="Embed content in a VAC's vector store",
                    inputSchema={
                        "type": "object",
                        "properties": {
                            "vac_name": {
                                "type": "string",
                                "description": "Name of the VAC to embed content for"
                            },
                            "content": {
                                "type": "string",
                                "description": "Content to embed"
                            },
                            "local_chunks": {
                                "type": "boolean",
                                "description": "Whether to process chunks locally",
                                "default": False
                            }
                        },
                        "required": ["vac_name", "content"]
                    }
                )
            ]

        @self.server.call_tool()
        async def call_tool(
            name: str,
            arguments: Any
        ) -> Sequence[TextContent | ImageContent | EmbeddedResource]:
            """Handle tool calls for Sunholo interactions"""
            
            if name == "chat_with_vac":
                if not isinstance(arguments, dict):
                    raise ValueError("Invalid arguments format")

                vac_name = arguments.get("vac_name")
                message = arguments.get("message")

                if not vac_name or not message:
                    raise ValueError("Missing required arguments")

                try:
                    cmd = ["sunholo", "vac", "chat", vac_name, message]
                    cmd.append("--headless")

                    result = subprocess.run(
                        cmd,
                        capture_output=True,
                        text=True
                    )

                    return [
                        TextContent(
                            type="text",
                            text=result.stdout
                        )
                    ]
                except subprocess.CalledProcessError as e:
                    return [
                        TextContent(
                            type="text",
                            text=f"Error chatting with VAC: {e.stderr}"
                        )
                    ]

            elif name == "embed_content":
                if not isinstance(arguments, dict):
                    raise ValueError("Invalid arguments format")

                vac_name = arguments.get("vac_name")
                content = arguments.get("content")
                local_chunks = arguments.get("local_chunks", False)

                if not vac_name or not content:
                    raise ValueError("Missing required arguments")

                try:
                    cmd = ["sunholo", "embed", vac_name, content]
                    if local_chunks:
                        cmd.append("--local-chunks")

                    result = subprocess.run(
                        cmd,
                        capture_output=True,
                        text=True
                    )

                    return [
                        TextContent(
                            type="text",
                            text=result.stdout
                        )
                    ]
                except subprocess.CalledProcessError as e:
                    return [
                        TextContent(
                            type="text",
                            text=f"Error embedding content: {e.stderr}"
                        )
                    ]
            elif name == "list_configs":

                # Build command
                cmd = ["sunholo", "list-configs"]
                
                if arguments.get("kind"):
                    cmd.extend(["--kind", arguments["kind"]])
                if arguments.get("vac"):
                    cmd.extend(["--vac", arguments["vac"]])
                if arguments.get("validate"):
                    cmd.append("--validate")

                # Execute sunholo command
                try:
                    result = subprocess.run(cmd, capture_output=True, text=True)
                    return [TextContent(type="text", text=result.stdout)]
                except subprocess.CalledProcessError as e:
                    return [TextContent(type="text", text=f"Error: {e.stderr}")]


            raise ValueError(f"Unknown tool: {name}")

    async def run(self):
        """Run the MCP server"""
        async with stdio_server() as (read_stream, write_stream):
            await self.server.run(
                read_stream,
                write_stream,
                self.server.create_initialization_options()
            )

def cli_mcp(args):
    """CLI handler for the MCP server command"""
    try:

        if not os.getenv("VAC_CONFIG_FOLDER"):
            raise ValueError("sunholo configuration folder must be present in config/ or via VAC_CONFIG_FOLDER")
        
        # Create and run the MCP server
        server = SunholoMCPServer()
        msg = {"message": "Starting Sunholo MCP server..."}

        logger.info(msg)
        console.print(msg)
        
        asyncio.run(server.run())

    except Exception as e:
        logger.error(f"Error running MCP server: {str(e)}")
        raise

def cli_mcp_bridge(args):
    """CLI handler for the MCP bridge command"""
    try:
        from .stdio_http_bridge import run_bridge
        
        http_url = args.url
        logger.info(f"Starting MCP stdio-to-HTTP bridge for {http_url}")
        
        if console:
            console.print(f"Starting MCP bridge to {http_url}...")
        
        asyncio.run(run_bridge(http_url))
        
    except KeyboardInterrupt:
        logger.info("Bridge stopped by user")
    except Exception as e:
        logger.error(f"Error running MCP bridge: {str(e)}")
        raise

def setup_mcp_subparser(subparsers):
    """
    Sets up an argparse subparser for the 'mcp' command 3.

    By default will use configurations within the folder specified by 'VAC_CONFIG_FOLDER'

    Example command:
    ```bash
    sunholo mcp
    ```
    """
    mcp_parser = subparsers.add_parser('mcp', 
                                      help='MCP (Model Context Protocol) commands')
    
    # Create subcommands for mcp
    mcp_subparsers = mcp_parser.add_subparsers(title='mcp commands',
                                               description='MCP subcommands',
                                               help='MCP subcommands',
                                               dest='mcp_command')
    
    # mcp server command (default)
    server_parser = mcp_subparsers.add_parser('server',
                                             help='Start an Anthropic MCP server that wraps sunholo functionality')
    server_parser.set_defaults(func=cli_mcp)
    
    # mcp bridge command
    bridge_parser = mcp_subparsers.add_parser('bridge',
                                             help='Start a stdio-to-HTTP bridge for MCP servers')
    bridge_parser.add_argument('url',
                                  nargs='?',
                                  default='http://127.0.0.1:1956/mcp',
                                  help='HTTP URL of the MCP server (default: http://127.0.0.1:1956/mcp)')
    bridge_parser.set_defaults(func=cli_mcp_bridge)
    
    # Set default behavior when no subcommand is provided
    mcp_parser.set_defaults(func=cli_mcp)
